# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Kapture to openmvg export functions.
"""

import json
import logging
import os
import os.path as path
from tqdm import tqdm
from typing import Dict, List, Optional, Union

import numpy as np
import quaternion

# kapture
import kapture
from kapture.core.Trajectories import rigs_remove_inplace
import kapture.io.csv
from kapture.io.binary import TransferAction, transfer_files_from_dir, array_to_file
from kapture.io.features import keypoints_to_filepaths, image_keypoints_from_file, get_keypoints_fullpath
from kapture.io.features import descriptors_to_filepaths, image_descriptors_from_file
from kapture.io.features import matches_to_filepaths, image_matches_from_file
from kapture.io.records import get_image_fullpath
import kapture.io.structure
from kapture.io.tar import TarCollection
from kapture.utils.Collections import try_get_only_key_from_collection
from kapture.utils.paths import safe_remove_file, safe_remove_any_path
# local
from .openmvg_commons import JSON_KEY, OPENMVG_SFM_DATA_VERSION_NUMBER, OPENMVG_DEFAULT_JSON_FILE_NAME
from .openmvg_commons import CameraModel

logger = logging.getLogger('openmvg')  # Using global openmvg logger

NEW_ID_MASK = 1 << 31  # 10000000 00000000 00000000 00000000
DEFAULT_FOCAL_LENGTH_FACTOR = 1.2


class CerealPointerRegistry:
    NULL_ID = 1 << 30  # 01000000 00000000 00000000 00000000

    def __init__(self, value_key, id_key):
        self._value_key = value_key
        self._id_key = id_key
        self._ids = {}
        self._id_current = 1

    def get_ids_dict(self, value):
        """ :return either id if known or id + name if new """
        if isinstance(value, dict) or value not in self._ids:
            # if this is the first time or a dict
            new_id = self._id_current
            if not isinstance(value, dict):
                self._ids[value] = new_id
            self._id_current += 1
            return {
                self._id_key: new_id | NEW_ID_MASK,
                self._value_key: value,
            }
        else:
            return {
                self._id_key: self._ids[value]
            }


def _get_camera_common_data(camera_params: list) -> Dict:
    # w, h, f, cx, cy
    data: Dict[str, Union[int, float, List[float]]] = {JSON_KEY.WIDTH: int(camera_params[0]),
                                                       JSON_KEY.HEIGHT: int(camera_params[1]),
                                                       JSON_KEY.FOCAL_LENGTH: float(camera_params[2]),
                                                       JSON_KEY.PRINCIPAL_POINT: [float(camera_params[3]),
                                                                                  float(camera_params[4])]}
    return data


def _get_intrinsic_pinhole(camera_params: list) -> Dict:
    # w, h, f, cx, cy
    return _get_camera_common_data(camera_params)


def _get_intrinsic_pinhole_radial_k1(camera_params: list) -> Dict:
    data = _get_camera_common_data(camera_params)
    # w, h, f, cx, cy, k
    data[JSON_KEY.DISTO_K1] = [float(camera_params[5])]
    return data


def _get_intrinsic_pinhole_radial_k3(camera_params: list) -> Dict:
    data = _get_camera_common_data(camera_params)
    # w, h, f, cx, cy, k1, k2, k3
    data[JSON_KEY.DISTO_K3] = [float(camera_params[5]), float(camera_params[6]), float(camera_params[7])]
    return data


def _get_intrinsic_pinhole_brown_t2(camera_params: list) -> Dict:
    # w, h, f, cx, cy, k1, k2, k3, t1, t2
    disto_t2 = [float(camera_params[5]), float(camera_params[6]), float(camera_params[7]),
                float(camera_params[8]), float(camera_params[9])]
    return {JSON_KEY.VALUE0: _get_camera_common_data(camera_params), JSON_KEY.DISTO_T2: disto_t2}


def _get_intrinsic_fisheye(camera_params: list) -> Dict:
    # w, h, f, cx, cy, k1, k2, k3, k4
    fisheye = [float(camera_params[5]), float(camera_params[6]), float(camera_params[7]), float(camera_params[8])]
    return {JSON_KEY.VALUE0: _get_camera_common_data(camera_params),
            JSON_KEY.FISHEYE: fisheye}


def _get_openmvg_id(kapture_id: str, kapture_to_openmvg_ids: Dict[str, int]) -> int:
    """ Get a valid openmvg id for the given kapture id.
    It keeps kapture_to_openmvg_ids up to date, and ensure there is no collision.

    :return: openmvg id
    """

    if kapture_id in kapture_to_openmvg_ids:
        # already defined
        return kapture_to_openmvg_ids[kapture_id]

    # its not defined, then, make up one
    last_known_openmvg_id = max(list(kapture_to_openmvg_ids.values())) if len(kapture_to_openmvg_ids) > 0 else -1
    assert type(last_known_openmvg_id) == int
    openmvg_id = last_known_openmvg_id + 1
    assert openmvg_id not in kapture_to_openmvg_ids.values()
    kapture_to_openmvg_ids[kapture_id] = openmvg_id
    return openmvg_id


def _get_openmvg_image_path(kapture_image_name: str, flatten_path: bool = False):
    """ the openmvg image sub path corresponding to the given kapture one. """
    return kapture_image_name if not flatten_path else kapture_image_name.replace('/', '_')


def _export_openmvg_intrinsics(
        kapture_cameras,
        kapture_to_openmvg_cam_ids: Dict[str, int],
        polymorphic_registry: CerealPointerRegistry,
        ptr_wrapper_registry: CerealPointerRegistry,
) -> List:
    """
    Exports the given kapture cameras to the openMVG sfm_data structure.
    In openMVG, cameras are referred as Intrinsics camera internal parameters.

    :param kapture_cameras: input kapture cameras to be exported (even if not used).
    :param kapture_to_openmvg_cam_ids: input/output dict that maps kapture camera ids to openMVG camera ids.
    :param polymorphic_registry: input/output polymorphic IDs status
    :param ptr_wrapper_registry: input/output polymorphic IDs status
    :return: intrinsics to be serialized
    """
    openmvg_intrinsics = []
    # process all cameras
    for kapture_cam_id, kapture_camera in kapture_cameras.items():
        openmvg_camera_id = kapture_to_openmvg_cam_ids.get(kapture_cam_id)
        if openmvg_camera_id is None:
            # this cameras is not used, skip it to make openMVG happy
            openmvg_camera_id = _get_openmvg_id(kapture_cam_id, kapture_to_openmvg_cam_ids)

        kapture_cam_type = kapture_camera.camera_type
        kapture_camera_params = kapture_camera.camera_params
        if kapture_cam_type == kapture.CameraType.SIMPLE_PINHOLE:
            # w, h, f, cx, cy
            openmvg_cam_type = CameraModel.pinhole
            data = _get_intrinsic_pinhole(kapture_camera_params)
        elif kapture_cam_type == kapture.CameraType.PINHOLE:
            # w, h, f, cx, cy
            openmvg_cam_type = CameraModel.pinhole
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            (kapture_camera_params[2] + kapture_camera_params[3]) / 2,  # fx+fy/2 as f
                            kapture_camera_params[4], kapture_camera_params[5]]  # cx cy
            data = _get_intrinsic_pinhole(faked_params)
        elif kapture_cam_type == kapture.CameraType.SIMPLE_RADIAL:
            # w, h, f, cx, cy, k
            openmvg_cam_type = CameraModel.pinhole_radial_k1
            data = _get_intrinsic_pinhole_radial_k1(kapture_camera_params)
        elif kapture_cam_type == kapture.CameraType.RADIAL:
            # w, h, f, cx, cy, k1, k2, k3
            openmvg_cam_type = CameraModel.pinhole_radial_k3
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            kapture_camera_params[2],  # f
                            kapture_camera_params[3], kapture_camera_params[4],  # cx cy
                            kapture_camera_params[5], kapture_camera_params[6], 0  # k1, k2, k3
                            ]
            data = _get_intrinsic_pinhole_radial_k3(faked_params)
        elif kapture_cam_type == kapture.CameraType.FULL_OPENCV or kapture_cam_type == kapture.CameraType.OPENCV:
            # w, h, f, cx, cy, k1, k2, k3, t1, t2
            openmvg_cam_type = CameraModel.pinhole_brown_t2
            k3 = kapture_camera_params[10] if len(kapture_camera_params) > 10 else 0
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            (kapture_camera_params[2] + kapture_camera_params[3]) / 2,  # fx+fy/2 as f
                            kapture_camera_params[4], kapture_camera_params[5],  # cx cy
                            kapture_camera_params[6], kapture_camera_params[7], k3,  # k1, k2, k3
                            kapture_camera_params[8], kapture_camera_params[9]  # p1, p2 (=t1, t2)
                            ]
            data = _get_intrinsic_pinhole_brown_t2(faked_params)
        elif kapture_cam_type == kapture.CameraType.OPENCV_FISHEYE:
            logger.warning('OpenCV fisheye model is not compatible with OpenMVG. Forcing distortion to 0')
            # w, h, f, cx, cy, k1, k2, k3, k4
            openmvg_cam_type = CameraModel.fisheye
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            (kapture_camera_params[2] + kapture_camera_params[3]) / 2,  # fx+fy/2 as f
                            kapture_camera_params[4], kapture_camera_params[5],  # cx cy
                            0, 0,  # k1, k2
                            0, 0  # k3, k4
                            ]
            data = _get_intrinsic_fisheye(faked_params)
        elif kapture_cam_type == kapture.CameraType.RADIAL_FISHEYE or \
                kapture_cam_type == kapture.CameraType.SIMPLE_RADIAL_FISHEYE:
            logger.warning('OpenCV fisheye model is not compatible with OpenMVG. Forcing distortion to 0')
            # w, h, f, cx, cy, k1, k2, k3, k4
            openmvg_cam_type = CameraModel.fisheye
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            kapture_camera_params[2],  # f
                            kapture_camera_params[3], kapture_camera_params[4],  # cx cy
                            0, 0,  # k1, k2
                            0, 0  # k3, k4
                            ]
            data = _get_intrinsic_fisheye(faked_params)
        elif kapture_cam_type == kapture.CameraType.UNKNOWN_CAMERA:
            logger.info(f'Camera {kapture_cam_id}: Unknown camera model, using simple radial')
            # Choose simple radial model, to allow openMVG to determine distortion param
            # w, h, f, cx, cy, k
            openmvg_cam_type = CameraModel.pinhole_radial_k1
            faked_params = [kapture_camera_params[0], kapture_camera_params[1],  # width height
                            max(kapture_camera_params[0], kapture_camera_params[1]) * DEFAULT_FOCAL_LENGTH_FACTOR,
                            # max(w,h)*1.2 as f
                            int(kapture_camera_params[0] / 2), int(kapture_camera_params[1] / 2),  # cx cy
                            0.0]  # k1
            data = _get_intrinsic_pinhole_radial_k1(faked_params)
        else:
            raise ValueError(f'Camera model {kapture_cam_type.value} not supported')

        intrinsic = {}
        cam_type_poly_id = polymorphic_registry.get_ids_dict(openmvg_cam_type.name)
        intrinsic.update(cam_type_poly_id)

        # it is assumed that this camera is only encountered once
        # set the first bit of ptr_wrapper_id_current to 1
        data_wrapper = ptr_wrapper_registry.get_ids_dict(data)

        intrinsic[JSON_KEY.PTR_WRAPPER] = data_wrapper
        openmvg_intrinsics.append({JSON_KEY.KEY: openmvg_camera_id, JSON_KEY.VALUE: intrinsic})

    return openmvg_intrinsics


def _export_openmvg_views(
        kapture_cameras: Dict[str, kapture.Camera],
        kapture_images: kapture.RecordsCamera,
        kapture_trajectories: kapture.Trajectories,
        kapture_to_openmvg_cam_ids: Dict[str, int],
        kapture_to_openmvg_view_ids: Dict[str, int],
        polymorphic_registry: CerealPointerRegistry,
        ptr_wrapper_registry: CerealPointerRegistry,
        image_path_flatten: bool,
) -> List:
    """

    :param kapture_cameras:
    :param kapture_images:
    :param kapture_trajectories:
    :param kapture_to_openmvg_cam_ids: input dict that maps kapture camera ids to openMVG camera ids.
    :param kapture_to_openmvg_view_ids: input dict that maps kapture image names to openMVG view ids.
    :param polymorphic_registry: input/output polymorphic IDs status
    :param ptr_wrapper_registry: input/output polymorphic IDs status
    :param image_path_flatten: flatten image path (eg. to avoid image name collision in openMVG regions).
    :return: views to be serialized
    """
    views = []

    """
    fills an array of views like :  
    {
            "key": 0,
            "value": {
                "polymorphic_id": 2147483650, "polymorphic_name": "view_priors", "ptr_wrapper": {
                    "id": 2147488881,
                    "data": {
                        "local_path": "db",
                        "filename": "1.jpg",
                        "width": 1600,
                        "height": 1063,
                        "id_view": 0,
                        "id_intrinsic": 3156,
                        "id_pose": 0,
                and optionally  
                        "use_pose_center_prior": true, "center_weight": [1.0,1.0,1.0], "center": [...],
                        "use_pose_rotation_prior": true, "rotation_weight": 1.0, "rotation": [[...],[...],[...]]
                    }
                }
            }
        },
    """
    # process all images
    for timestamp, kapture_cam_id, kapture_image_name in kapture.flatten(kapture_images):
        openmvg_cam_id = _get_openmvg_id(kapture_cam_id, kapture_to_openmvg_cam_ids)
        openmvg_view_id = _get_openmvg_id(kapture_image_name, kapture_to_openmvg_view_ids)
        openmvg_image_filepath = _get_openmvg_image_path(kapture_image_name, image_path_flatten)
        openmvg_image_filename = path.basename(openmvg_image_filepath)
        openmvg_image_local_path = path.dirname(openmvg_image_filepath)
        kapture_camera_params = kapture_cameras[kapture_cam_id].camera_params
        view_data = {JSON_KEY.LOCAL_PATH: openmvg_image_local_path,
                     JSON_KEY.FILENAME: openmvg_image_filename,
                     JSON_KEY.WIDTH: int(kapture_camera_params[0]),
                     JSON_KEY.HEIGHT: int(kapture_camera_params[1]),
                     JSON_KEY.ID_VIEW: openmvg_view_id,
                     JSON_KEY.ID_INTRINSIC: openmvg_cam_id,
                     JSON_KEY.ID_POSE: openmvg_view_id}

        view = {}
        # retrieve image pose from trajectories
        if timestamp not in kapture_trajectories:
            view[JSON_KEY.POLYMORPHIC_ID] = CerealPointerRegistry.NULL_ID
        else:
            # there is a pose for that timestamp
            # The poses are stored both as priors (in the 'views' table) and as known poses (in the 'extrinsics' table)
            assert kapture_cam_id in kapture_trajectories[timestamp]
            view_priors_id = polymorphic_registry.get_ids_dict(JSON_KEY.VIEW_PRIORS)
            view.update(view_priors_id)

            pose_tr = kapture_trajectories[timestamp].get(kapture_cam_id)
            prior_q = pose_tr.r
            prior_t = pose_tr.inverse().t_raw
            pose_data = {JSON_KEY.CENTER: prior_t,
                         JSON_KEY.ROTATION: quaternion.as_rotation_matrix(prior_q).tolist()}

            view_data[JSON_KEY.USE_POSE_CENTER_PRIOR] = True
            view_data[JSON_KEY.CENTER_WEIGHT] = [1.0, 1.0, 1.0]
            view_data[JSON_KEY.CENTER] = prior_t
            view_data[JSON_KEY.USE_POSE_ROTATION_PRIOR] = True
            view_data[JSON_KEY.ROTATION_WEIGHT] = 1.0
            view_data[JSON_KEY.ROTATION] = pose_data[JSON_KEY.ROTATION]

        # it is assumed that this view is only encountered once
        view_wrapper = ptr_wrapper_registry.get_ids_dict(view_data)
        view[JSON_KEY.PTR_WRAPPER] = view_wrapper
        views.append({JSON_KEY.KEY: openmvg_view_id, JSON_KEY.VALUE: view})

    return views


def _export_openmvg_extrinsics(
        kapture_images: kapture.RecordsCamera,
        kapture_trajectories: kapture.Trajectories,
        kapture_to_openmvg_view_ids: Dict[str, int],
) -> List:
    """

    :param kapture_images: all kapture images
    :param kapture_trajectories: all kapture poses
    :param kapture_to_openmvg_view_ids: dict that maps kapture image ids to openMVG view ids.
    :return: extrinsics to be serialized
    """
    extrinsics = []
    # process all images
    for timestamp, kapture_cam_id, kapture_image_name in kapture.flatten(kapture_images):
        openmvg_view_id = kapture_to_openmvg_view_ids.get(kapture_image_name)
        if openmvg_view_id is None:
            # this pose corresponds to no views (orphan), openMVG does not want it.
            logger.debug(f'skipping extrinsic for {kapture_image_name} (not referenced in views).')
            continue

        # retrieve image pose from trajectories
        if timestamp in kapture_trajectories:
            # there is a pose for that timestamp
            # The poses are stored both as priors (in the 'views' table) and as known poses (in the 'extrinsics' table)
            assert kapture_cam_id in kapture_trajectories[timestamp]
            pose_tr = kapture_trajectories[timestamp].get(kapture_cam_id)
            prior_q = pose_tr.r
            prior_t = pose_tr.inverse().t_raw
            pose_data = {JSON_KEY.CENTER: prior_t,
                         JSON_KEY.ROTATION: quaternion.as_rotation_matrix(prior_q).tolist()}
            extrinsics.append({JSON_KEY.KEY: openmvg_view_id, JSON_KEY.VALUE: pose_data})

    return extrinsics


def _export_openmvg_structure(
        kapture_points_3d: Optional[kapture.Points3d],
        kapture_to_openmvg_view_ids: Dict[str, int],
        kapture_observations: Optional[kapture.Observations] = None,
        kapture_keypoints: Optional[kapture.Keypoints] = None,
        keypoints_type: Optional[str] = None,
        kapture_path: Optional[str] = None,
        tar_handlers: Optional[TarCollection] = None,
) -> Optional[List]:
    """
    Exports kapture 3D points, observations and key points

    :param kapture_points_3d: 3D points to export
    :param kapture_to_openmvg_view_ids: kapture to openmvg view identifiers
    :param kapture_observations: kapture observations if any
    :param kapture_keypoints: kapture key points if any
    :param keypoints_type: key points type if any
    :param kapture_path: path to kapture top directory
    :param tar_handlers: list of tar to use to read kapture data

    :return: openmvg structure to be serialized
    """
    # early check
    if kapture_points_3d is None:
        logger.warning('no 3D points to export.')
        return None

    xyz_coordinates = kapture_points_3d[:, 0:3]
    include_2d_observations = kapture_observations is not None
    openmvg_structure = []
    # this loop can be very long, lets show some progress
    hide_progress_bars = logger.getEffectiveLevel() > logging.INFO

    for point_idx, coords in enumerate(tqdm(xyz_coordinates, disable=hide_progress_bars)):
        point_3d_structure = {
            'key': point_idx,
            'value': {
                'X': coords.tolist(),
                'observations': []
            }
        }
        if include_2d_observations and \
                kapture_observations is not None and \
                point_idx in kapture_observations and \
                keypoints_type is not None and \
                keypoints_type in kapture_observations[point_idx]:
            for kapture_image_name, feature_point_id in kapture_observations[point_idx, keypoints_type]:
                openmvg_view_id = kapture_to_openmvg_view_ids[kapture_image_name]
                point_2d_observation = {'key': openmvg_view_id,
                                        'value': {'id_feat': feature_point_id, }}

                if kapture_path and kapture_keypoints is not None:
                    # if given, load keypoints to populate 2D coordinates of the feature.
                    keypoints_file_path = get_keypoints_fullpath(keypoints_type,
                                                                 kapture_path,
                                                                 kapture_image_name,
                                                                 tar_handlers)
                    try:
                        keypoints_data = image_keypoints_from_file(keypoints_file_path,
                                                                   dtype=kapture_keypoints.dtype,
                                                                   dsize=kapture_keypoints.dsize)
                        point_2d_observation['value']['x'] = keypoints_data[feature_point_id, 0:2].tolist()
                    except FileNotFoundError:
                        logger.warning(f'unable to load keypoints file {keypoints_file_path}')

                point_3d_structure['value']['observations'].append(point_2d_observation)

        openmvg_structure.append(point_3d_structure)

    return openmvg_structure


def _export_openmvg_sfm_data(
    kapture_data: kapture.Kapture,
    kapture_path: str,
    tar_handlers: TarCollection,
    keypoints_type: Optional[str],
    openmvg_sfm_data_file_path: str,
    openmvg_image_root_path: str,
    image_action: TransferAction,
    image_path_flatten: bool,
    force: bool,
    kapture_to_openmvg_view_ids: Dict
) -> None:
    """
    Convert the kapture data into an openMVG dataset stored as a dictionary, and store it as a json file.
    The format is defined here:
    https://openmvg.readthedocs.io/en/latest/software/SfM/SfM_OutputFormat/

    :param kapture_data: the kapture data
    :param kapture_path: top directory of the kapture data and the images
    :param tar_handlers: tar handlers to use to read the data
    :param keypoints_type: type of key points if any
    :param openmvg_sfm_data_file_path: path to the SfM data file to be written.
    :param openmvg_image_root_path: input path to openMVG image directory to be created.
    :param image_action: action to apply on images: link, copy, move or do nothing.
    :param image_path_flatten: flatten image path (eg. to avoid image name collision in openMVG regions).
    :param force: if true, will remove existing openMVG data without prompting the user.
    :param kapture_to_openmvg_view_ids: input/output mapping of kapture image name to corresponding openmvg view id.
    """

    if kapture_data.cameras is None or kapture_data.records_camera is None:
        raise ValueError('export_openmvg_sfm_data needs kapture camera and records_camera.')

    if image_action == TransferAction.root_link:
        raise NotImplementedError('root link is not implemented, use skip instead.')

    # refer to the original image dir when skipping image transfer.
    if image_action == TransferAction.skip:
        openmvg_image_root_path = get_image_fullpath(kapture_path)

    if openmvg_image_root_path is None:
        raise ValueError(f'openmvg_image_root_path must be defined to be able to perform {image_action}.')

    # make sure directory is ready to contain openmvg_sfm_data_file_path
    os.makedirs(path.dirname(openmvg_sfm_data_file_path), exist_ok=True)

    # Check we don't have other sensors defined
    if len(kapture_data.sensors) != len(kapture_data.cameras):
        extra_sensor_number = len(kapture_data.sensors) - len(kapture_data.cameras)
        logger.warning(f'We will ignore {extra_sensor_number} sensors that are not camera')

    # openmvg does not support rigs
    if kapture_data.rigs:
        logger.info('remove rigs notation.')
        rigs_remove_inplace(kapture_data.trajectories, kapture_data.rigs)
        kapture_data.rigs.clear()

    # Compute root path and camera used in records
    kapture_to_openmvg_cam_ids = {}  # kapture_cam_id -> openmvg_cam_id

    # polymorphic_status = PolymorphicStatus({}, 1, 1)
    polymorphic_registry = CerealPointerRegistry(id_key=JSON_KEY.POLYMORPHIC_ID, value_key=JSON_KEY.POLYMORPHIC_NAME)
    ptr_wrapper_registry = CerealPointerRegistry(id_key=JSON_KEY.ID, value_key=JSON_KEY.DATA)

    logger.debug('exporting intrinsics ...')
    openmvg_sfm_data_intrinsics = _export_openmvg_intrinsics(
        kapture_data.cameras,
        kapture_to_openmvg_cam_ids,
        polymorphic_registry,
        ptr_wrapper_registry
    )

    logger.debug('exporting views ...')
    openmvg_sfm_data_views = _export_openmvg_views(
        kapture_data.cameras,
        kapture_data.records_camera,
        kapture_data.trajectories,
        kapture_to_openmvg_cam_ids,
        kapture_to_openmvg_view_ids,
        polymorphic_registry,
        ptr_wrapper_registry,
        image_path_flatten
    )

    logger.debug('exporting extrinsics ...')
    openmvg_sfm_data_poses = _export_openmvg_extrinsics(
        kapture_data.records_camera,
        kapture_data.trajectories,
        kapture_to_openmvg_view_ids)

    # structure : correspond to kapture observations + 3D points
    logger.debug('exporting structure ...')
    kapture_keypoints = kapture_data.keypoints[keypoints_type] if (
            kapture_data.keypoints is not None and
            keypoints_type is not None and
            keypoints_type in kapture_data.keypoints) else None
    openmvg_sfm_data_structure = _export_openmvg_structure(
        kapture_data.points3d,
        kapture_to_openmvg_view_ids,
        kapture_data.observations,
        kapture_keypoints,
        keypoints_type,
        kapture_path,
        tar_handlers
    )

    openmvg_sfm_data = {
        JSON_KEY.SFM_DATA_VERSION: OPENMVG_SFM_DATA_VERSION_NUMBER,
        JSON_KEY.ROOT_PATH: path.abspath(openmvg_image_root_path),
        JSON_KEY.INTRINSICS: openmvg_sfm_data_intrinsics,
        JSON_KEY.VIEWS: openmvg_sfm_data_views,
        JSON_KEY.EXTRINSICS: openmvg_sfm_data_poses,
        JSON_KEY.STRUCTURE: openmvg_sfm_data_structure,
        JSON_KEY.CONTROL_POINTS: [],
    }

    logger.debug(f'Saving to openmvg {openmvg_sfm_data_file_path}...')
    with open(openmvg_sfm_data_file_path, "w") as fid:
        json.dump(openmvg_sfm_data, fid, indent=4)

    # do the actual image transfer
    if not image_action == TransferAction.skip:
        job_copy = (
            (  # source path -> destination path
                get_image_fullpath(kapture_path, kapture_image_name),
                path.join(openmvg_image_root_path, _get_openmvg_image_path(kapture_image_name, image_path_flatten))
            )
            for kapture_image_name in kapture_data.records_camera.data_list()
        )
        source_filepath_list, destination_filepath_list = zip(*job_copy)
        transfer_files_from_dir(source_filepath_list, destination_filepath_list, image_action, force)


def _export_openmvg_regions(
        kapture_path: str,
        kapture_keypoints: Optional[kapture.Keypoints],
        keypoints_type: Optional[str],
        kapture_descriptors: Optional[kapture.Descriptors],
        descriptors_type: Optional[str],
        tar_handlers: TarCollection,
        openmvg_regions_dir_path: str,
        image_path_flatten: bool
):
    """
    exports openMVG regions ie keypoints and descriptors.

    :param kapture_path: input path to root kapture directory.
    :param kapture_keypoints: input kapture keypoints. Could be None if no keypoints.
    :param keypoints_type: type of key points if any
    :param kapture_descriptors: input kapture descriptors. Could be None if no descriptors.
    :param descriptors_type: type of descriptors if any
    :param tar_handlers: tar handlers to read the data
    :param openmvg_regions_dir_path: input path to output openMVG regions directory.
    :param image_path_flatten: if true, it means that image path are to be flatten.
    """
    # early check we should do
    if kapture_keypoints is None or kapture_descriptors is None or keypoints_type is None or descriptors_type is None:
        logger.warning('no keypoints or descriptors to export.')
        return

    # make sure output directory is ready
    os.makedirs(openmvg_regions_dir_path, exist_ok=True)

    # only able to export SIFT
    if any([f.type_name.upper() != 'SIFT' for f in [kapture_keypoints, kapture_descriptors]]):
        raise ValueError(f'unable to export other regions than sift '
                         f'(got {kapture_keypoints.type_name}/{kapture_descriptors.type_name})')

    os.makedirs(openmvg_regions_dir_path, exist_ok=True)
    polymorphic_registry = CerealPointerRegistry(id_key=JSON_KEY.POLYMORPHIC_ID, value_key=JSON_KEY.POLYMORPHIC_NAME)
    # create image_describer.json
    fake_regions_type = {"ptr_wrapper": {"valid": 1, "data": {"value0": [], "value1": []}}}
    fake_regions_type.update(polymorphic_registry.get_ids_dict('SIFT_Regions'))
    image_describer = {
        'regions_type': fake_regions_type
    }
    image_describer_file_path = path.join(openmvg_regions_dir_path, 'image_describer.json')
    with open(image_describer_file_path, 'w') as fid:
        json.dump(image_describer, fid, indent=4)

    # this loop can be very long, lets show some progress
    hide_progress_bars = logger.getEffectiveLevel() > logging.INFO

    # copy keypoints files
    keypoints = keypoints_to_filepaths(kapture_keypoints, keypoints_type, kapture_path, tar_handlers)
    for kapture_image_name, kapture_keypoint_file_path in tqdm(keypoints.items(), disable=hide_progress_bars):
        openmvg_keypoint_file_name = _get_openmvg_image_path(kapture_image_name, image_path_flatten)
        openmvg_keypoint_file_name = path.splitext(path.basename(openmvg_keypoint_file_name))[0] + '.feat'
        openmvg_keypoint_file_path = path.join(openmvg_regions_dir_path, openmvg_keypoint_file_name)
        keypoints_data = image_keypoints_from_file(kapture_keypoint_file_path,
                                                   kapture_keypoints.dtype,
                                                   kapture_keypoints.dsize)
        keypoints_data = keypoints_data[:, 0:4]
        np.savetxt(openmvg_keypoint_file_path, keypoints_data, fmt='%10.5f')

    # copy descriptors files
    """
    from openMVG regions_factory.hpp
    using SIFT_Regions = Scalar_Regions<SIOPointFeature, unsigned char, 128>;
    using AKAZE_Float_Regions = Scalar_Regions<SIOPointFeature, float, 64>;
    using AKAZE_Liop_Regions = Scalar_Regions<SIOPointFeature, unsigned char, 144>;
    using AKAZE_Binary_Regions = Binary_Regions<SIOPointFeature, 64>;
    """
    descriptors = descriptors_to_filepaths(kapture_descriptors, descriptors_type, kapture_path, tar_handlers)
    for kapture_image_name, kapture_descriptors_file_path in tqdm(descriptors.items(), disable=hide_progress_bars):
        openmvg_descriptors_file_name = _get_openmvg_image_path(kapture_image_name, image_path_flatten)
        openmvg_descriptors_file_name = path.splitext(path.basename(openmvg_descriptors_file_name))[0] + '.desc'
        openmvg_descriptors_file_path = path.join(openmvg_regions_dir_path, openmvg_descriptors_file_name)
        kapture_descriptors_data = image_descriptors_from_file(kapture_descriptors_file_path,
                                                               kapture_descriptors.dtype,
                                                               kapture_descriptors.dsize)
        # assign a byte array of [size_t[1] + uint8[nb features x 128]
        size_t_len = 64 // 8
        openmvg_descriptors_data = np.empty(dtype=np.uint8, shape=(kapture_descriptors_data.size + size_t_len,))
        openmvg_descriptors_data[0:size_t_len].view(dtype=np.uint64)[0] = kapture_descriptors_data.shape[0]
        openmvg_descriptors_data[size_t_len:] = kapture_descriptors_data.flatten()
        array_to_file(openmvg_descriptors_file_path, openmvg_descriptors_data)


def _export_openmvg_matches(
        kapture_path: str,
        kapture_data: kapture.Kapture,
        keypoints_type: Optional[str],
        tar_handlers: TarCollection,
        openmvg_matches_file_path: str,
        kapture_to_openmvg_view_ids: Dict[str, int]
):
    if kapture_data.matches is None or keypoints_type is None or keypoints_type not in kapture_data.matches:
        logger.warning('No matches to be exported.')
        return

    if path.splitext(openmvg_matches_file_path)[1] != '.txt':
        logger.warning('Matches are exported as text format, even if file does not ends with .txt.')

    # make sure output directory is ready
    os.makedirs(path.dirname(openmvg_matches_file_path), exist_ok=True)

    hide_progress_bars = logger.getEffectiveLevel() > logging.INFO
    matches = matches_to_filepaths(kapture_data.matches[keypoints_type], keypoints_type, kapture_path, tar_handlers)
    with open(openmvg_matches_file_path, 'w') as fid:
        for image_pair, kapture_matches_filepath in tqdm(matches.items(), disable=hide_progress_bars):
            # idx image1 idx image 2
            # nb pairs
            # pl1 pr1 pl2 pr2 ...
            i, j = [kapture_to_openmvg_view_ids[image_name] for image_name in image_pair]
            fid.write(f'{i} {j}\n')
            matches_indices = image_matches_from_file(kapture_matches_filepath)[:, 0:2].astype(int)
            fid.write(f'{matches_indices.shape[0]}\n')
            for indices_pair in matches_indices:
                fid.write(f'{indices_pair[0]}  {indices_pair[1]}\n')


def export_openmvg(
    kapture_path: str,
    openmvg_sfm_data_file_path: str,
    openmvg_image_root_path: str = None,
    openmvg_regions_dir_path: str = None,
    openmvg_matches_file_path: str = None,
    image_action: TransferAction = TransferAction.skip,
    image_path_flatten: bool = False,
    keypoints_type: Optional[str] = None,
    descriptors_type: Optional[str] = None,
    force: bool = False
) -> None:
    """
    Export the kapture data to an openMVG files.
    If the openmvg_path is a directory, it will create a JSON file (using the default name sfm_data.json)
    in that directory.

    :param kapture_path: full path to input kapture directory
    :param openmvg_sfm_data_file_path: input path to the SfM data file to be written.
    :param openmvg_image_root_path: optional input path to openMVG image directory to be created.
    :param openmvg_regions_dir_path: optional input path to openMVG regions (feat, desc) directory to be created.
    :param openmvg_matches_file_path: optional input path to openMVG matches file to be created.
    :param image_action: an action to apply on the images: relative linking, absolute linking, copy or move. Or top
     directory linking or skip to do nothing. If not "skip" equires openmvg_image_root_path to be defined.
    :param image_path_flatten: flatten image path (eg. to avoid image name collision in openMVG regions).
    :param force: if true, will remove existing openMVG data without prompting the user.
    :param keypoints_type: key points type if any
    :param descriptors_type: descriptors type if any
    """

    if any(arg is not None and not isinstance(arg, str)
           for arg in [kapture_path, openmvg_image_root_path, openmvg_regions_dir_path, openmvg_matches_file_path]
           ):
        raise ValueError('expect str (or None) as path argument.')

    # clean before export
    safe_remove_file(openmvg_sfm_data_file_path, force)
    if path.exists(openmvg_sfm_data_file_path):
        raise ValueError(f'{openmvg_sfm_data_file_path} file already exist')

    # load kapture
    logger.info(f'loading kapture {kapture_path}...')
    with kapture.io.csv.get_all_tar_handlers(kapture_path) as tar_handlers:
        kapture_data = kapture.io.csv.kapture_from_dir(kapture_path, tar_handlers=tar_handlers)
        if kapture_data is None or not isinstance(kapture_data, kapture.Kapture):
            raise ValueError(f'unable to load kapture from {kapture_path}')
        kapture_to_openmvg_view_ids = {}

        if keypoints_type is None:
            keypoints_type: str = try_get_only_key_from_collection(kapture_data.keypoints)
        if descriptors_type is None:
            descriptors_type: str = try_get_only_key_from_collection(kapture_data.descriptors)

        logger.info(f'exporting sfm data to {openmvg_sfm_data_file_path} ...')
        _export_openmvg_sfm_data(
            kapture_data,
            kapture_path,
            tar_handlers,
            keypoints_type,
            openmvg_sfm_data_file_path,
            openmvg_image_root_path,
            image_action,
            image_path_flatten,
            force,
            kapture_to_openmvg_view_ids)

        if openmvg_regions_dir_path is not None:
            try:
                logger.info(f'exporting regions to {openmvg_regions_dir_path} ...')
                kapture_keypoints = kapture_data.keypoints[keypoints_type] if (
                        kapture_data.keypoints is not None and
                        keypoints_type is not None and
                        keypoints_type in kapture_data.keypoints) else None
                kapture_descriptors = kapture_data.descriptors[descriptors_type] if (
                        kapture_data.descriptors is not None and
                        descriptors_type is not None and
                        descriptors_type in kapture_data.descriptors) else None
                _export_openmvg_regions(
                    kapture_path,
                    kapture_keypoints,
                    keypoints_type,
                    kapture_descriptors,
                    descriptors_type,
                    tar_handlers,
                    openmvg_regions_dir_path,
                    image_path_flatten
                )
            except ValueError as e:
                logger.error(e)

        if openmvg_matches_file_path is not None:
            try:
                logger.info(f'exporting matches to {openmvg_matches_file_path} ...')
                _export_openmvg_matches(
                    kapture_path,
                    kapture_data,
                    keypoints_type,
                    tar_handlers,
                    openmvg_matches_file_path,
                    kapture_to_openmvg_view_ids
                )
            except ValueError as e:
                logger.error(e)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Openmvg to kapture import functions.
"""

import json
import logging
import quaternion
import numpy as np
import os
import os.path as path
import shutil
from tqdm import tqdm
from typing import Dict, Optional, Union
# kapture
import kapture
import kapture.io.csv as kcsv
import kapture.io.structure
from kapture.io.records import TransferAction, get_image_fullpath
from kapture.io.features import get_keypoints_fullpath, get_descriptors_fullpath
from kapture.io.features import get_matches_fullpath
from kapture.io.binary import array_to_file
from kapture.utils.paths import path_secure
# local
from .openmvg_commons import JSON_KEY, OPENMVG_DEFAULT_JSON_FILE_NAME
from .openmvg_commons import CameraModel

logger = logging.getLogger('openmvg')  # Using global openmvg logger


def import_openmvg(
        sfm_data_path: str,
        regions_dir_path: Optional[str],
        matches_file_path: Optional[str],
        kapture_path: str,
        image_action: TransferAction,
        force_overwrite_existing: bool = False) -> None:
    """
    Converts an openMVG JSON file to a kapture directory.
    If an image action is provided (link, copy or move), links to the image files are created,
    or the image files are copied or moved.

    :param sfm_data_path: path to the openMVG sfm_data file.
    :param regions_dir_path: input path to directory containing regions (*.feat, *.desc)
    :param matches_file_path: input path to openMVG matches file (eg. matches.f.bin)
    :param kapture_path: path to the kapture directory where the data will be exported
    :param image_action: action to apply to the images
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    """

    if path.isdir(sfm_data_path):
        sfm_data_path = path.join(sfm_data_path, OPENMVG_DEFAULT_JSON_FILE_NAME)
    # sanity check
    if not path.isfile(sfm_data_path):
        raise ValueError(f'OpenMVG JSON file {sfm_data_path} does not exist.')

    os.makedirs(kapture_path, exist_ok=True)
    kapture.io.structure.delete_existing_kapture_files(kapture_path, force_overwrite_existing)

    logger.info(f'Loading sfm_data file {sfm_data_path}')
    with open(sfm_data_path, 'r') as f:
        input_json = json.load(f)
        kapture_data = import_openmvg_sfm_data_json(input_json, kapture_path, image_action)

    if regions_dir_path:
        logger.info(f'Loading regions from {regions_dir_path}')
        _import_openmvg_regions(regions_dir_path, kapture_data, kapture_path)

    if matches_file_path:
        logger.info(f'Loading matches from {matches_file_path}')
        _import_openmvg_matches(matches_file_path, kapture_data, kapture_path)

    logger.info(f'Saving to kapture {kapture_path}')
    kcsv.kapture_to_dir(kapture_path, kapture_data)


GET_ID_MASK = 2147483647  # 01111111 11111111 11111111 11111111
ID_POSE_NOT_LOCALIZED = 4294967295  # 11111111 11111111 11111111 11111111


def import_openmvg_sfm_data_json(
        sfm_data_json: Dict[str, Union[str, Dict]],
        kapture_images_path: str,
        image_action=TransferAction.skip) -> kapture.Kapture:
    """
    Imports an openMVG sfm_data json structure to a kapture object.
    Also copy, move or link the images files if necessary.

    :param sfm_data_json: the openmvg JSON parsed as a dictionary
    :param kapture_images_path: top directory to create the kapture images tree
    :param image_action: action to apply on images: link, copy, move or do nothing.
    :return: the constructed kapture object
    """

    data_root_path: str = ''

    if sfm_data_json[JSON_KEY.ROOT_PATH]:
        data_root_path = sfm_data_json[JSON_KEY.ROOT_PATH]
    elif image_action == TransferAction.skip:
        logger.warning('No root_path in sfm_data.')
    else:  # It is needed to execute an action with the image file
        raise ValueError(f"Missing root_path to do image action '{image_action.name}'")
    openmvg_images_dir = path.basename(data_root_path)

    # Imports all the data from the json file to kapture objects
    kapture_cameras = _import_openmvg_cameras(sfm_data_json)
    device_identifiers = {int: str}  # Pose id -> device id
    timestamp_for_pose = {int: int}  # Pose id -> timestamp
    # Imports the images as records_camera, but also fill in the devices_identifiers and timestamp_for_pose dictionaries
    records_camera = _import_openmvg_images(
        sfm_data_json, image_action, kapture_images_path, openmvg_images_dir, data_root_path,
        device_identifiers, timestamp_for_pose)
    trajectories = _import_openmvg_trajectories(
        sfm_data_json, device_identifiers, timestamp_for_pose)

    kapture_data = kapture.Kapture(sensors=kapture_cameras, records_camera=records_camera, trajectories=trajectories)
    return kapture_data


def _import_openmvg_cameras(input_json) -> kapture.Sensors:  # noqa: C901
    kapture_cameras = kapture.Sensors()
    if input_json.get(JSON_KEY.INTRINSICS):
        polymorphic_id_to_value = {}
        logger.info('Importing intrinsics')
        for sensor in input_json[JSON_KEY.INTRINSICS]:
            value = sensor[JSON_KEY.VALUE]
            if JSON_KEY.POLYMORPHIC_NAME in value:
                # new type name: store it for next instances
                polymorphic_id = value[JSON_KEY.POLYMORPHIC_ID] & GET_ID_MASK
                polymorphic_id_to_value[polymorphic_id] = value[JSON_KEY.POLYMORPHIC_NAME]
                logger.debug("New camera_type: " + polymorphic_id_to_value[polymorphic_id])
            else:
                if JSON_KEY.POLYMORPHIC_ID not in value:
                    raise ValueError(f'{JSON_KEY.POLYMORPHIC_ID} is missing (intrinsics)')
                polymorphic_id = value[JSON_KEY.POLYMORPHIC_ID]

            if polymorphic_id not in polymorphic_id_to_value:
                raise ValueError(f'Unknown polymorphic_id {polymorphic_id}')

            camera_model = CameraModel(polymorphic_id_to_value[polymorphic_id])
            camera_data = value[JSON_KEY.PTR_WRAPPER][JSON_KEY.DATA]

            if camera_model == CameraModel.pinhole:
                # w, h, f, cx, cy
                camera = kapture.Camera(kapture.CameraType.SIMPLE_PINHOLE, [
                    int(camera_data[JSON_KEY.WIDTH]),
                    int(camera_data[JSON_KEY.HEIGHT]),
                    camera_data[JSON_KEY.FOCAL_LENGTH],
                    camera_data[JSON_KEY.PRINCIPAL_POINT][0],
                    camera_data[JSON_KEY.PRINCIPAL_POINT][1],
                ])
            elif camera_model == CameraModel.pinhole_radial_k1:
                # w, h, f, cx, cy, k
                camera = kapture.Camera(kapture.CameraType.SIMPLE_RADIAL, [
                    int(camera_data[JSON_KEY.WIDTH]),
                    int(camera_data[JSON_KEY.HEIGHT]),
                    camera_data[JSON_KEY.FOCAL_LENGTH],
                    camera_data[JSON_KEY.PRINCIPAL_POINT][0],
                    camera_data[JSON_KEY.PRINCIPAL_POINT][1],
                    camera_data[JSON_KEY.DISTO_K1][0]
                ])
            elif camera_model == CameraModel.pinhole_radial_k3:
                # w, h, f, cx, cy, k1, k2, k3
                camera = kapture.Camera(kapture.CameraType.RADIAL, [
                    int(camera_data[JSON_KEY.WIDTH]),
                    int(camera_data[JSON_KEY.HEIGHT]),
                    camera_data[JSON_KEY.FOCAL_LENGTH],
                    camera_data[JSON_KEY.PRINCIPAL_POINT][0],
                    camera_data[JSON_KEY.PRINCIPAL_POINT][1],
                    camera_data[JSON_KEY.DISTO_K3][0],
                    camera_data[JSON_KEY.DISTO_K3][1]
                ])
                # camera_data["disto_k3"][2] ignored: radial model has two distortion param, while openMVG's has three
            elif camera_model == CameraModel.pinhole_brown_t2:
                # w, h, f, cx, cy, k1, k2, k3, t1, t2
                if float(camera_data[JSON_KEY.DISTO_T2][2]) != 0:
                    # if k3 not null, use FULL_OPENCV, otherwise OPENCV
                    # w, h, fx, fy, cx, cy, k1, k2, p1, p2, k3, k4, k5, k6
                    value0 = camera_data[JSON_KEY.VALUE0]
                    disto_t2 = camera_data[JSON_KEY.DISTO_T2]
                    camera = kapture.Camera(kapture.CameraType.FULL_OPENCV, [
                        int(value0[JSON_KEY.WIDTH]),
                        int(value0[JSON_KEY.HEIGHT]),
                        value0[JSON_KEY.FOCAL_LENGTH],
                        value0[JSON_KEY.FOCAL_LENGTH],
                        value0[JSON_KEY.PRINCIPAL_POINT][0],
                        value0[JSON_KEY.PRINCIPAL_POINT][1],
                        disto_t2[0], disto_t2[1], disto_t2[3], disto_t2[4], disto_t2[2],
                        0, 0, 0
                    ])
                else:
                    # w, h, fx, fy, cx, cy, k1, k2, p1, p2
                    value0 = camera_data[JSON_KEY.VALUE0]
                    disto_t2 = camera_data[JSON_KEY.DISTO_T2]
                    camera = kapture.Camera(kapture.CameraType.OPENCV, [
                        int(value0[JSON_KEY.WIDTH]),
                        int(value0[JSON_KEY.HEIGHT]),
                        value0[JSON_KEY.FOCAL_LENGTH],
                        value0[JSON_KEY.FOCAL_LENGTH],
                        value0[JSON_KEY.PRINCIPAL_POINT][0],
                        value0[JSON_KEY.PRINCIPAL_POINT][1],
                        disto_t2[0], disto_t2[1], disto_t2[3], disto_t2[4]])
            elif camera_model == CameraModel.fisheye:
                logger.warning(
                    "OpenMVG fisheye models are not compatible with OpenCV."
                    " Using SIMPLE_RADIAL_FISHEYE and forcing distortion to 0")
                # w, h, f, cx, cy, k
                value0 = camera_data[JSON_KEY.VALUE0]
                camera = kapture.Camera(kapture.CameraType.SIMPLE_RADIAL_FISHEYE, [
                    int(value0[JSON_KEY.WIDTH]),
                    int(value0[JSON_KEY.HEIGHT]),
                    value0[JSON_KEY.FOCAL_LENGTH],
                    value0[JSON_KEY.PRINCIPAL_POINT][0],
                    value0[JSON_KEY.PRINCIPAL_POINT][1],
                    0])
            else:
                raise ValueError(f'Camera model {camera_model} not supported')

            kapture_cameras[str(sensor[JSON_KEY.KEY])] = camera

    return kapture_cameras


def _import_openmvg_images(input_json, image_action, kapture_images_path, openmvg_images_dir, root_path,
                           device_identifiers, timestamp_for_pose):
    records_camera = kapture.RecordsCamera()
    if input_json.get(JSON_KEY.VIEWS):
        views = input_json[JSON_KEY.VIEWS]
        if image_action == TransferAction.root_link:
            # Do a unique images directory link
            # kapture/<records_dir>/openmvg_top_images_directory -> openmvg_root_path
            kapture_records_path = get_image_fullpath(kapture_images_path)
            os.makedirs(kapture_records_path, exist_ok=True)
            os.symlink(root_path, path.join(kapture_records_path, openmvg_images_dir))
        logger.info(f'Importing {len(views)} images')
        # Progress bar only in debug or info level
        if image_action != TransferAction.skip and image_action != TransferAction.root_link \
                and logger.getEffectiveLevel() <= logging.INFO:
            progress_bar = tqdm(total=len(views))
        else:
            progress_bar = None
        for view in views:
            input_data = view[JSON_KEY.VALUE][JSON_KEY.PTR_WRAPPER][JSON_KEY.DATA]
            pose_id = input_data[JSON_KEY.ID_POSE]
            # All two values should be the same (?)
            if input_data[JSON_KEY.ID_VIEW]:
                timestamp = input_data[JSON_KEY.ID_VIEW]
            else:
                timestamp = view[JSON_KEY.KEY]
            device_id = str(input_data[JSON_KEY.ID_INTRINSIC])  # device_id must be a string for kapture
            device_identifiers[pose_id] = device_id
            timestamp_for_pose[pose_id] = timestamp

            kapture_filename = _import_openmvg_image_file(input_data, openmvg_images_dir, root_path,
                                                          kapture_images_path, image_action)

            progress_bar and progress_bar.update(1)

            key = (timestamp, device_id)  # tuple of int,str
            records_camera[key] = path_secure(kapture_filename)
        progress_bar and progress_bar.close()
    return records_camera


def _import_openmvg_image_file(input_data, openmvg_images_dir, root_path, kapture_images_path, image_action) -> str:
    # Add the common openmvg images directory in front of the filename
    filename: str
    if input_data.get(JSON_KEY.LOCAL_PATH):
        filename = path.join(input_data[JSON_KEY.LOCAL_PATH], input_data[JSON_KEY.FILENAME])
    else:
        filename = input_data[JSON_KEY.FILENAME]
    kapture_filename = path.join(openmvg_images_dir, filename)
    if image_action != TransferAction.skip and image_action != TransferAction.root_link:
        src_path: str
        if root_path:
            src_path = path.join(root_path, filename)
        else:
            src_path = filename
        dst_path = get_image_fullpath(kapture_images_path, kapture_filename)
        # Create destination directory if necessary
        dst_dir = path.dirname(dst_path)
        if not path.isdir(dst_dir):
            os.makedirs(dst_dir, exist_ok=True)
        # Check if already exist
        if path.exists(dst_path):
            os.unlink(dst_path)
        # Create file or link
        if image_action == TransferAction.copy:
            shutil.copy2(src_path, dst_path)
        elif image_action == TransferAction.move:
            shutil.move(src_path, dst_path)
        else:
            # Individual link
            if image_action == TransferAction.link_relative:
                # Compute relative path
                src_path = path.relpath(src_path, dst_dir)
            os.symlink(src_path, dst_path)
            # Symlink might crash on Windows if the user executing this code has no admin privilege
    return kapture_filename


def _import_openmvg_trajectories(input_json, device_identifiers, timestamp_for_pose):
    trajectories = kapture.Trajectories()
    if input_json.get(JSON_KEY.EXTRINSICS):
        extrinsics = input_json[JSON_KEY.EXTRINSICS]
        logger.info(f'Importing {len(extrinsics)} extrinsics -> trajectories')
        for pose in extrinsics:
            pose_id = pose[JSON_KEY.KEY]
            center = pose[JSON_KEY.VALUE][JSON_KEY.CENTER]
            rotation = pose[JSON_KEY.VALUE][JSON_KEY.ROTATION]
            kap_translation = -1 * np.matmul(rotation, center)
            kap_pose = kapture.PoseTransform(quaternion.from_rotation_matrix(rotation), kap_translation)
            timestamp = timestamp_for_pose.get(pose_id)
            if timestamp is None:
                logger.warning(f'Missing timestamp for extrinsic {pose_id}')
                continue
            device_id = device_identifiers.get(pose_id)
            if device_id is None:
                logger.warning(f'Missing device for extrinsic {pose_id}')
                continue
            trajectories[(timestamp, device_id)] = kap_pose  # tuple of int,str -> 6D pose
    return trajectories


def _import_openmvg_regions(
        openmvg_regions_directory_path,
        kapture_data,
        kapture_path
):
    # look for the "image_describer.json"
    image_describer_path = path.join(openmvg_regions_directory_path, 'image_describer.json')
    if not path.isfile(image_describer_path):
        logger.debug(f'file not found : {image_describer_path}')
        return

    with open(image_describer_path) as f:
        image_describer = json.load(f)

    # retrieve what type of keypoints it is.
    keypoints_type = image_describer.get('regions_type', {}).get('polymorphic_name', 'UNDEFINED')
    keypoints_name = {
        'SIFT_Regions': 'SIFT',
        'AKAZE_Float_Regions': 'AKAZE'
    }.get(keypoints_type, keypoints_type)
    kapture_keypoints = kapture.Keypoints(type_name=keypoints_name, dtype=float, dsize=4)

    # retrieve what type of descriptors it is.
    descriptors_type = image_describer.get('image_describer', {}).get('polymorphic_name', 'UNDEFINED')
    descriptors_props = {
        'SIFT_Image_describer': dict(type_name='SIFT', dtype=np.int32, dsize=128,
                                     keypoints_type=keypoints_type,
                                     metric_type='L2'),
        'AKAZE_Image_describer_SURF': dict(type_name='AKAZE', dtype=np.int32, dsize=128,
                                           keypoints_type=keypoints_type,
                                           metric_type='L2'),
    }.get(descriptors_type)
    if not descriptors_props:
        raise ValueError(f'conversion of {descriptors_type} descriptors not implemented.')
    kapture_descriptors = kapture.Descriptors(**descriptors_props)

    # populate regions files in openMVG directory
    # https://github.com/openMVG/openMVG/blob/master/src/openMVG/features/scalar_regions.hpp#L23
    image_names = kapture_data.records_camera.data_list()
    for image_name in image_names:
        openmvg_image_name = path.splitext(path.basename(image_name))[0]
        # keypoints
        keypoints_data = None
        openmvg_keypoints_filepath = path.join(openmvg_regions_directory_path, openmvg_image_name + '.feat')
        if path.isfile(openmvg_keypoints_filepath):
            # there is a keypoints file in openMVG, lets add it to kapture
            keypoints_data = np.loadtxt(openmvg_keypoints_filepath)
            assert keypoints_data.shape[1] == 4
            kapture_keypoints.add(image_name)
            # and convert file
            kapture_keypoints_filepath = get_keypoints_fullpath(keypoints_type,
                                                                kapture_path,
                                                                image_name)
            array_to_file(kapture_keypoints_filepath, keypoints_data)

        # descriptors
        openmvg_descriptors_filepath = path.join(openmvg_regions_directory_path, openmvg_image_name + '.desc')
        if path.isfile(openmvg_descriptors_filepath):
            assert path.isfile(openmvg_keypoints_filepath)
            # there is a keypoints file in openMVG, lets add it to kapture
            # assumes descriptors shape from keypoints_data shape
            descriptors_data_bytes = np.fromfile(openmvg_descriptors_filepath, dtype=np.uint8)
            nb_features = keypoints_data.shape[0] if keypoints_data is not None else 0
            descriptors_shape = descriptors_data_bytes[0:8].view(descriptors_props['dtype'])
            assert descriptors_shape[0] == nb_features
            descriptors_data = descriptors_data_bytes[8:].view(np.uint8).reshape((nb_features, 128))
            # descriptors_data.reshape((keypoints_data.shape[0], -1))
            kapture_descriptors.add(image_name)
            # and convert file
            kapture_descriptors_filepath = get_descriptors_fullpath(descriptors_type,
                                                                    kapture_path,
                                                                    image_name)
            array_to_file(kapture_descriptors_filepath, descriptors_data)

    kapture_data.keypoints = {keypoints_type: kapture_keypoints}
    kapture_data.descriptors = {descriptors_type: kapture_descriptors}


def _import_openmvg_matches(
        matches_file_path: str,
        kapture_data: kapture.Kapture,
        kapture_path: str):
    if kapture_data.records_camera is None:
        logger.warning('no images in records_camera, cannot import matches')
        return
    if kapture_data.keypoints is None:
        logger.warning('no keypoints, cannot import matches')
        return
    assert len(kapture_data.keypoints) == 1
    keypoints_type = next(iter(kapture_data.keypoints.keys()))

    # idx image1 idx image 2
    # nb pairs
    # pl1 pr1 pl2 pr2 ...
    openmvg_image_idx_to_kapture_image_name = {}
    matches = kapture.Matches()
    with open(matches_file_path, 'r') as fid:
        while True:
            line = fid.readline()
            if not line:
                break
            splits_idx = line.rstrip('\r\n').split()
            assert len(splits_idx) == 2
            idx_image1 = int(splits_idx[0])
            if idx_image1 in openmvg_image_idx_to_kapture_image_name:
                image_1 = openmvg_image_idx_to_kapture_image_name[idx_image1]
            else:
                if idx_image1 not in kapture_data.records_camera:
                    raise ValueError(f'{idx_image1} not in kapture_data.records_camera')
                assert len(kapture_data.records_camera[idx_image1]) == 1
                sensor_id = next(iter(kapture_data.records_camera[idx_image1].keys()))
                image_1 = kapture_data.records_camera.get(idx_image1)[sensor_id]
                openmvg_image_idx_to_kapture_image_name[idx_image1] = image_1

            idx_image2 = int(splits_idx[1])
            if idx_image2 in openmvg_image_idx_to_kapture_image_name:
                image_2 = openmvg_image_idx_to_kapture_image_name[idx_image2]
            else:
                if idx_image2 not in kapture_data.records_camera:
                    raise ValueError(f'{idx_image2} not in kapture_data.records_camera')
                assert len(kapture_data.records_camera[idx_image2]) == 1
                sensor_id = next(iter(kapture_data.records_camera[idx_image2].keys()))
                image_2 = kapture_data.records_camera.get(idx_image2)[sensor_id]
                openmvg_image_idx_to_kapture_image_name[idx_image2] = image_2

            swap_order = image_2 < image_1
            line = fid.readline()
            num_matches = int(line.rstrip('\r\n'))
            matches_array = np.empty((num_matches, 3), dtype=np.float)
            for i in range(num_matches):
                line = fid.readline()
                splits_kpts_idx = line.rstrip('\r\n').split()
                assert len(splits_kpts_idx) == 2
                if swap_order:
                    matches_array[i, 1] = int(splits_kpts_idx[0])
                    matches_array[i, 0] = int(splits_kpts_idx[1])
                else:
                    matches_array[i, 0] = int(splits_kpts_idx[0])
                    matches_array[i, 1] = int(splits_kpts_idx[1])
                matches_array[i, 2] = 1.0
            if swap_order:
                image_filename_pair = (image_2, image_1)
                matches.add(image_2, image_1)
            else:
                image_filename_pair = (image_1, image_2)
                matches.add(image_1, image_2)
            matches_filepath = get_matches_fullpath(image_filename_pair, keypoints_type, kapture_path)
            array_to_file(matches_filepath, matches_array)
    kapture_data.matches = {keypoints_type: matches}



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script exports data from kapture to COLMAP database and/or reconstruction files.
"""

import logging
import os
import os.path as path
from typing import Optional
import warnings

# kapture
import kapture
from kapture.core.Trajectories import rigs_remove_inplace
import kapture.algo.merge_keep_ids
import kapture.io.csv as csv
import kapture.io.structure
import kapture.utils.paths
from kapture.utils.Collections import try_get_only_key_from_collection

# local
from .database import COLMAPDatabase
from .database_extra import is_colmap_db_empty, kapture_to_colmap, get_colmap_image_ids_from_db,\
    get_colmap_camera_ids_from_db
from .export_colmap_rigs import export_colmap_rig
from .export_colmap_reconstruction import export_to_colmap_txt

logger = logging.getLogger('colmap')


def export_colmap(kapture_dir_path: str,
                  colmap_database_filepath: str,
                  colmap_reconstruction_dir_path: Optional[str],
                  keypoints_type: Optional[str] = None,
                  descriptors_type: Optional[str] = None,
                  colmap_rig_filepath: str = None,
                  force_overwrite_existing: bool = False,
                  pairsfile_path: Optional[str] = None) -> None:
    """
    Exports kapture data to colmap database and or reconstruction text files.

    :param kapture_dir_path: kapture top directory
    :param colmap_database_filepath: path to colmap database file
    :param colmap_reconstruction_dir_path: path to colmap reconstruction directory
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    :param descriptors_type: type of descriptors to export, name of the descriptors subfolder
    :param colmap_rig_filepath: path to colmap rig file
    :param force_overwrite_existing: Silently overwrite colmap files if already exists.
    :param pairsfile_path: Filter matches to load
    """

    os.makedirs(path.dirname(colmap_database_filepath), exist_ok=True)
    if colmap_reconstruction_dir_path:
        os.makedirs(colmap_reconstruction_dir_path, exist_ok=True)

    assert colmap_database_filepath
    kapture.utils.paths.safe_remove_file(colmap_database_filepath, force_overwrite_existing, 'colmap database')

    logger.info('creating colmap database in {}'.format(colmap_database_filepath))
    db = COLMAPDatabase.connect(colmap_database_filepath)
    if not is_colmap_db_empty(db):
        raise ValueError('the existing colmap database is not empty : {}'.format(colmap_database_filepath))

    logger.info('loading kapture files...')
    with csv.get_all_tar_handlers(kapture_dir_path) as tar_handlers:
        kapture_data = csv.kapture_from_dir(kapture_dir_path, pairsfile_path, tar_handlers=tar_handlers)
        assert isinstance(kapture_data, kapture.Kapture)

        if keypoints_type is None:
            keypoints_type = try_get_only_key_from_collection(kapture_data.keypoints)
        # if it's still None try again from matches
        if keypoints_type is None:
            keypoints_type = try_get_only_key_from_collection(kapture_data.matches)
        if descriptors_type is None:
            descriptors_type = try_get_only_key_from_collection(kapture_data.descriptors)

        # COLMAP does not fully support rigs.
        if kapture_data.rigs is not None:
            # make sure, rigs are not used in trajectories.
            logger.info('remove rigs notation.')
            rigs_remove_inplace(kapture_data.trajectories, kapture_data.rigs)

        # write colmap database
        kapture_to_colmap(kapture_data, kapture_dir_path, tar_handlers, db, keypoints_type, descriptors_type)

        if colmap_reconstruction_dir_path:
            # create text files
            colmap_camera_ids = get_colmap_camera_ids_from_db(db, kapture_data.records_camera)
            colmap_image_ids = get_colmap_image_ids_from_db(db)
            export_to_colmap_txt(colmap_reconstruction_dir_path,
                                 kapture_data,
                                 kapture_dir_path,
                                 tar_handlers,
                                 colmap_camera_ids,
                                 colmap_image_ids,
                                 keypoints_type)
            if colmap_rig_filepath:
                try:
                    if kapture_data.rigs is None:
                        raise ValueError('No rig to export.')
                    export_colmap_rig(colmap_rig_filepath,
                                      kapture_data.rigs,
                                      kapture_data.records_camera,
                                      colmap_camera_ids)
                except Exception as e:
                    warnings.warn(e)
                    logger.warning(e)

    db.close()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Operations on kapture pose objects
"""

import math
import numpy as np
import quaternion
from typing import Tuple, Union, List

import kapture


def pose_transform_distance(pose_a: kapture.PoseTransform, pose_b: kapture.PoseTransform) -> Tuple[float, float]:
    """
    get translation and rotation distance between two PoseTransform

    :return: (position_distance, rotation_distance in rad), can be nan is case of invalid comparison
    """
    # handle NoneType with try expect blocks
    try:
        translation_distance = np.linalg.norm(pose_a.t - pose_b.t)
    except TypeError:
        translation_distance = math.nan

    try:
        rotation_distance = quaternion.rotation_intrinsic_distance(pose_a.r, pose_b.r)
    except TypeError:
        rotation_distance = math.nan
    return translation_distance, rotation_distance


def world_pose_transform_distance(pose_a: kapture.PoseTransform, pose_b: kapture.PoseTransform) -> Tuple[float, float]:
    """
    get position and rotation error between two PoseTransform
    pose_a and pose_b should be world to device

    :return: (position_distance, rotation_distance in deg), can be nan is case of invalid comparison
    """
    if pose_a.r is None and pose_a.t is None:
        return math.nan, math.nan
    elif pose_a.r is None:
        device_to_world_a = kapture.PoseTransform(r=None, t=-pose_a.t)
    elif pose_a.t is None:
        device_to_world_a = kapture.PoseTransform(r=pose_a.r.inverse(), t=None)
    else:
        device_to_world_a = pose_a.inverse()

    if pose_b.r is None and pose_b.t is None:
        return math.nan, math.nan
    elif pose_b.r is None:
        device_to_world_b = kapture.PoseTransform(r=None, t=-pose_b.t)
    elif pose_b.t is None:
        device_to_world_b = kapture.PoseTransform(r=pose_b.r.inverse(), t=None)
    else:
        device_to_world_b = pose_b.inverse()

    pose_error = pose_transform_distance(device_to_world_a, device_to_world_b)
    return pose_error[0], np.rad2deg(pose_error[1])


def average_quaternion(big_q: np.ndarray) -> np.ndarray:
    """
    Computes the Chordal L2-Mean using quaternions.
    Ported from Tolga Birdal's implementation
    https://github.com/tolgabirdal/averaging_quaternions/blob/master/avg_quaternion_markley.m (MIT)

    :param big_q: Q (or big_q in python) is an (M,4) ndarray of quaternions
    :return: float array representing the average quaternion
    """
    # Form the symmetric accumulator matrix
    big_a = np.zeros((4, 4))
    big_m = big_q.shape[0]
    for i in range(big_m):
        q = big_q[i, :]
        big_a += np.outer(q, q)  # rank 1 update
    # scale
    big_a /= big_m
    # Get the eigen vector corresponding to largest eigen value
    return np.linalg.eigh(big_a)[1][:, -1]


def average_pose_transform(poses: List[kapture.PoseTransform]) -> kapture.PoseTransform:
    """
    average a list of poses with equal weights

    :param poses: list of poses to average
    :return: average PoseTransform
    """
    assert isinstance(poses, list)
    assert len(poses) > 0

    # handle NoneType with try expect blocks
    try:
        translation = np.sum(tuple([pose.t for pose in poses]), axis=0) / len(poses)
    except TypeError:
        translation = None

    try:
        rotation = average_quaternion(np.vstack(tuple([quaternion.as_float_array(pose.r) for pose in poses])))
    except TypeError:
        rotation = None

    return kapture.PoseTransform(r=rotation, t=translation)


def average_quaternion_weighted(big_q: np.ndarray, weights: Union[List[float], np.ndarray]) -> np.ndarray:
    """
    Averaging Quaternions.
    Ported from Tolga Birdal's implementation
    https://github.com/tolgabirdal/averaging_quaternions/blob/master/wavg_quaternion_markley.m (MIT)

    :param big_q: Q is an (M,4) ndarray of quaternions
    :param weights: a (M,) vector
    :return: float array representing the average quaternion
    """
    # Form the symmetric accumulator matrix
    big_a = np.zeros((4, 4))
    big_m = big_q.shape[0]
    w_sum = 0

    for i in range(big_m):
        q = big_q[i, :]
        w_i = weights[i]
        big_a += w_i * (np.outer(q, q))  # rank 1 update
        w_sum += w_i
    # scale
    big_a /= w_sum
    # Get the eigen vector corresponding to largest eigen value
    return np.linalg.eigh(big_a)[1][:, -1]


def average_pose_transform_weighted(poses: List[kapture.PoseTransform],
                                    weights: Union[List[float], np.ndarray]) -> kapture.PoseTransform:
    """
    average a list of poses with any weights

    :param poses: list of poses to average
    :param weights: a (len(poses),) vector
    :return: average PoseTransform
    """

    assert isinstance(poses, list)
    assert len(poses) > 0

    assert isinstance(weights, list)
    assert len(weights) == len(poses)

    weight_sum = np.sum(weights)

    # handle NoneType with try expect blocks
    try:
        translation = np.sum(tuple([weights[i] * pose.t for i, pose in enumerate(poses)]), axis=0) / weight_sum
    except TypeError:
        translation = None

    try:
        rotation = average_quaternion_weighted(np.vstack(tuple([quaternion.as_float_array(pose.r) for pose in poses])),
                                               weights)
    except TypeError:
        rotation = None

    return kapture.PoseTransform(r=rotation, t=translation)



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script exports kapture data to a COLMAP database and/or reconstruction files
"""

import argparse
import logging
import sys
import path_to_kapture  # enables import kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.converter.colmap.export_colmap import export_colmap


logger = logging.getLogger('colmap')


def colmap_command_line():
    """
    Do the export of kapture data using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(description='exports kapture data to Colmap database and/or text files.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete colmap if already exists.')
    # create the parser for the export command #########################################################################
    parser.add_argument('-i', '-k', '--kapture', required=True,
                        help='input path to kapture data root directory')
    parser.add_argument('--pairsfile-path', default=None, type=str,
                        help=('text file in the csv format; where each line is image_name1, image_name2, score '
                              'which contains the matches to export'))
    parser.add_argument('-kpt', '--keypoints-type', default=None, help='kapture keypoints type.')
    parser.add_argument('-desc', '--descriptors-type', default=None, help='kapture descriptors type.')
    parser.add_argument('-db', '--database', required=True,
                        help='database output path.')
    parser.add_argument('-txt', '--reconstruction',
                        help='text reconstruction output path.')
    parser.add_argument('-rig', '--rig',
                        help='json rig output path.')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # for debug, let kapture express itself.
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    logger.debug(f'{sys.argv[0]} \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()))

    logger.info('exporting colmap ...')
    export_colmap(args.kapture, args.database, args.reconstruction,
                  args.keypoints_type, args.descriptors_type,
                  args.rig, args.force, args.pairsfile_path)
    logger.info('done.')


if __name__ == '__main__':
    colmap_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to import a 7scenes model into a kapture.
The RGB-D Dataset 7-Scenes data structure is defined here:
    https://www.microsoft.com/en-us/research/project/rgb-d-dataset-7-scenes/

    Each sequence (seq-XX.zip) consists of 500-1000 frames. Each frame consists of three files:
     - Color: frame-XXXXXX.color.png (RGB, 24-bit, PNG)
     - Depth: frame-XXXXXX.depth.png (depth in millimeters, 16-bit, PNG, invalid depth is set to 65535).
     - Pose: frame-XXXXXX.pose.txt (camera-to-world, 4×4 matrix in homogeneous coordinates).

"""

import argparse
import logging
import os
import os.path as path
import re
import numpy as np
import quaternion
from typing import Optional
from PIL import Image
from tqdm import tqdm
# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
import kapture.io.features
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from kapture.utils.paths import path_secure

logger = logging.getLogger('7scenes')
MODEL = kapture.CameraType.RADIAL

POSE_SUFFIX = 'pose'
RGB_SUFFIX = 'color'
DEPTH_SUFFIX = 'depth'
RGB_SENSOR_ID = 'kinect_rgb'
DEPTH_SENSOR_ID = 'kinect_depth'
REG_DEPTH_SENSOR_ID = 'kinect_depth_reg'
RGBD_SENSOR_ID = 'kinect'
PARTITION_FILENAMES = {
    'mapping': 'TrainSplit.txt',
    'query': 'TestSplit.txt'
}


def get_camera_matrix(fx: float, fy: float, cx: float, cy: float):
    return np.array([[fx, 0, cx],
                     [0, fy, cy],
                     [0, 0, 1]])


def get_K(camera_type: kapture.CameraType, camera_params):
    if camera_type == kapture.CameraType.UNKNOWN_CAMERA:
        width = camera_params[0]
        height = camera_params[1]
        focal = 1.2 * max(width, height)
        return get_camera_matrix(focal, focal, camera_params[0] / 2, camera_params[1] / 2)
    elif camera_type in [kapture.CameraType.PINHOLE, kapture.CameraType.OPENCV,
                         kapture.CameraType.OPENCV_FISHEYE, kapture.CameraType.FULL_OPENCV,
                         kapture.CameraType.FOV, kapture.CameraType.THIN_PRISM_FISHEYE]:
        return get_camera_matrix(camera_params[2], camera_params[3], camera_params[4], camera_params[5])
    else:
        return get_camera_matrix(camera_params[2], camera_params[3], camera_params[4], camera_params[4])


def register_depth(Kdepth, Kcolor, Rt, depth, width_color, height_color):
    reg_depth = np.zeros((height_color, width_color), dtype=np.float32)
    y, x = np.meshgrid(np.arange(depth.shape[0]), np.arange(depth.shape[1]), indexing='ij')
    x = x.reshape(1, -1)
    y = y.reshape(1, -1)
    z = depth.reshape(1, -1)
    x = (x - Kdepth[0, 2]) / Kdepth[0, 0]
    y = (y - Kdepth[1, 2]) / Kdepth[1, 1]
    pts = np.vstack((x * z, y * z, z))
    pts = Rt[:3, :3] @ pts + Rt[:3, 3:]
    pts = Kcolor @ pts
    px = np.round(pts[0, :] / pts[2, :])
    py = np.round(pts[1, :] / pts[2, :])
    mask = (px >= 0) * (py >= 0) * (px < width_color) * (py < height_color)
    reg_depth[py[mask].astype(int), px[mask].astype(int)] = pts[2, mask]
    return reg_depth


def import_7scenes(d7scenes_path: str,
                   kapture_dir_path: str,
                   force_overwrite_existing: bool = False,
                   images_import_method: TransferAction = TransferAction.skip,
                   partition: Optional[str] = None
                   ) -> None:
    """
    Imports RGB-D Dataset 7-Scenes dataset and save them as kapture.

    :param d7scenes_path: path to the 7scenes sequence root path
    :param kapture_dir_path: path to kapture top directory
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files.
    :param partition: if specified = 'mapping' or 'query'. Requires d7scenes_path/TestSplit.txt or TrainSplit.txt
                    to exists.
    """
    os.makedirs(kapture_dir_path, exist_ok=True)
    delete_existing_kapture_files(kapture_dir_path, force_erase=force_overwrite_existing)

    logger.info('loading all content ...')

    d7s_filename_re = re.compile(r'((?P<sequence>.+)/)?frame-(?P<frame_id>\d{6})\.(?P<suffix>\w*)\.(?P<ext>\w*)')

    # populate all relevant files
    d7s_filenames = (path_secure(path.relpath(path.join(dp, fn), d7scenes_path))
                     for dp, _, fs in os.walk(d7scenes_path) for fn in fs)

    logger.info('populating 7-scenes files ...')
    d7s_filenames = {filename: d7s_filename_re.search(filename).groupdict()
                     for filename in sorted(d7s_filenames)
                     if d7s_filename_re.search(filename)}

    # reorg as shot[seq, id] = {color: , depth: , pose: , ...}
    shots = {}
    for timestamp, (filename, file_attribs) in enumerate(d7s_filenames.items()):
        shot_id = (file_attribs.get('sequence'), file_attribs['frame_id'])
        shots.setdefault(shot_id, {})[file_attribs['suffix']] = filename

    # fake timestamps
    for timestamp, shot_id in enumerate(shots):
        shots[shot_id]['timestamp'] = timestamp

    # if given, filter partition
    if partition is not None:
        # read the authors split file
        partition_filepath = path.join(d7scenes_path, PARTITION_FILENAMES[partition])
        if not path.isfile(partition_filepath):
            raise FileNotFoundError(f'partition file is missing: {partition_filepath}.')
        with open(partition_filepath, 'rt') as file:
            split_sequences = [f'seq-{int(seq.strip()[len("sequence"):]):02}' for seq in file.readlines()]
        assert len(split_sequences) > 0
        # filter out
        shots = {(seq, frame): shot
                 for (seq, frame), shot in shots.items()
                 if seq in split_sequences}

    if len(shots) == 0:
        raise FileNotFoundError('no file found: make sure the path to 7scenes sequence is valid.')

    # eg. shots['seq-01', '000000'] =
    #       {
    #           'color': 'seq-01/frame-000000.color.jpg',
    #           'depth': 'seq-01/frame-000000.depth.png',
    #           'pose': 'seq-01/frame-000000.pose.txt',
    #           'timestamp': 0}

    # images + depth maps
    logger.info('populating image and depth maps files ...')
    snapshots = kapture.RecordsCamera()
    depth_maps = kapture.RecordsDepth()
    for shot in shots.values():
        snapshots[shot['timestamp'], RGB_SENSOR_ID] = shot['color']
        kapture_depth_map_filename = shot['depth'][:-len('.png')]  # kapture depth files are not png
        depth_maps[shot['timestamp'], DEPTH_SENSOR_ID] = kapture_depth_map_filename
        kapture_registered_depth_map_filename = shot['depth'][:-len('.png')] + '.reg'  # kapture depth files are not png
        depth_maps[shot['timestamp'], REG_DEPTH_SENSOR_ID] = kapture_registered_depth_map_filename

    # poses
    logger.info('import poses files ...')
    trajectories = kapture.Trajectories()
    for shot in shots.values():
        pose_filepath = path.join(d7scenes_path, shot['pose'])
        pose_mat = np.loadtxt(pose_filepath)  # camera-to-world, 4×4 matrix in homogeneous coordinates
        rotation_mat = pose_mat[0:3, 0:3]
        position_vec = pose_mat[0:3, 3]
        rotation_quat = quaternion.from_rotation_matrix(rotation_mat)
        pose_world_from_cam = kapture.PoseTransform(r=rotation_quat, t=position_vec)
        pose_cam_from_world = pose_world_from_cam.inverse()
        trajectories[shot['timestamp'], RGBD_SENSOR_ID] = pose_cam_from_world

    # sensors
    """
    From authors: The RGB and depth camera have not been calibrated and we can’t provide calibration parameters at the
    moment. The recorded frames correspond to the raw, uncalibrated camera images. In the KinectFusion pipeline we used
    the following default intrinsics for the depth camera: Principle point (320,240), Focal length (585,585).
    ----
    We use the extr. kinect camera parameters from
    https://projet.liris.cnrs.fr/voir/activities-dataset/kinect-calibration.html.
    """
    sensors = kapture.Sensors()
    # camera_type = kapture.CameraType.OPENCV
    # # w, h, f, cx, cy, k1, k2, p1, p2, k3
    # camera_params = [640, 480, 5.2161910696979987e+02,
    #                  5.2132946256749767e+02, 3.1755491910920682e+02,
    #                  2.5921654718027673e+02, 2.5673002693536984e-01,
    #                  -9.3976085633794137e-01, -1.8605549188751580e-03, -2.2232238578189420e-03]
    camera_type = kapture.CameraType.SIMPLE_PINHOLE
    # w, h, fx, fy, cx, cy
    # camera_params = [640, 480,
    #                  5.2161910696979987e+02, 5.2132946256749767e+02,
    #                  3.1755491910920682e+02, 2.5921654718027673e+02]
    camera_params = [640, 480, 525, 320, 240]
    sensors[RGB_SENSOR_ID] = kapture.Camera(
        name=RGB_SENSOR_ID,
        camera_type=camera_type,
        camera_params=camera_params
    )
    # depth_camera_type = kapture.CameraType.OPENCV
    # w, h, f, cx, cy, k1, k2, p1, p2, k3
    # depth_camera_params = [640, 480, 5.8818670481438744e+02,
    #                        5.8724220649505514e+02, 3.1076280589210484e+02,
    #                        2.2887144980135292e+02, -1.8932947734719333e-01,
    #                        1.1358015104098631e+00, -4.4260345347128536e-03,
    #                        -5.4869578635708153e-03, -2.2460143607712921e+00]
    depth_camera_type = kapture.CameraType.SIMPLE_PINHOLE
    # w, h, fx, fy, cx, cy
    # depth_camera_params = [640, 480,
    #                        5.8818670481438744e+02, 5.8724220649505514e+02,
    #                        3.1076280589210484e+02, 2.2887144980135292e+02]
    # w, h, f, cx, cy
    depth_camera_params = [640, 480, 585, 320, 240]
    sensors[DEPTH_SENSOR_ID] = kapture.Camera(
        name=DEPTH_SENSOR_ID,
        camera_type=depth_camera_type,
        camera_params=depth_camera_params,
        sensor_type='depth'
    )
    sensors[REG_DEPTH_SENSOR_ID] = kapture.Camera(
        name=REG_DEPTH_SENSOR_ID,
        camera_type=depth_camera_type,
        camera_params=camera_params,
        sensor_type='depth'
    )

    # bind camera and depth sensor into a rig
    R = np.array([[9.9996518012567637e-01, 2.6765126468950343e-03, -7.9041012313000904e-03],
                  [-2.7409311281316700e-03, 9.9996302803027592e-01, -8.1504520778013286e-03],
                  [7.8819942130445332e-03, 8.1718328771890631e-03, 9.9993554558014031e-01]])
    T = np.array([-2.5558943178152542e-02, 1.0109636268061706e-04, 2.0318321729487039e-03])
    Rt = np.vstack((np.hstack((R, T.reshape(3, 1))), np.array([0, 0, 0, 1])))
    logger.info('building rig with camera and depth sensor ...')
    rigs = kapture.Rigs()
    rigs[RGBD_SENSOR_ID, RGB_SENSOR_ID] = kapture.PoseTransform(quaternion.from_rotation_matrix(R), T)
    rigs[RGBD_SENSOR_ID, REG_DEPTH_SENSOR_ID] = kapture.PoseTransform(quaternion.from_rotation_matrix(R), T)
    rigs[RGBD_SENSOR_ID, DEPTH_SENSOR_ID] = kapture.PoseTransform()

    # import (copy) image files.
    logger.info('copying image files ...')
    image_filenames = [f for _, _, f in kapture.flatten(snapshots)]
    import_record_data_from_dir_auto(d7scenes_path, kapture_dir_path, image_filenames, images_import_method)

    # import (copy) depth map files.
    logger.info('converting depth files ...')
    depth_map_filenames = kapture.io.records.records_to_filepaths(depth_maps, kapture_dir_path)
    hide_progress = logger.getEffectiveLevel() > logging.INFO
    for depth_map_filename, depth_map_filepath_kapture in tqdm(depth_map_filenames.items(), disable=hide_progress):
        if '.reg' in depth_map_filename:
            continue
        depth_map_filepath_7scenes = path.join(d7scenes_path, depth_map_filename + '.png')
        depth_map = np.array(Image.open(depth_map_filepath_7scenes))
        # change invalid depth from 65535 to 0
        depth_map[depth_map == 65535] = 0
        # depth maps is in mm in 7scenes, convert it to meters
        depth_map = depth_map.astype(np.float32) * 1.0e-3
        kapture.io.records.records_depth_to_file(depth_map_filepath_kapture, depth_map)
        # register depth to rgb
        reg_depth_map = register_depth(get_K(depth_camera_type, depth_camera_params), get_K(camera_type, camera_params),
                                       Rt, depth_map, camera_params[0], camera_params[1])
        kapture.io.records.records_depth_to_file(depth_map_filepath_kapture + '.reg', reg_depth_map)

    # pack into kapture format
    imported_kapture = kapture.Kapture(
        records_camera=snapshots,
        records_depth=depth_maps,
        rigs=rigs,
        trajectories=trajectories,
        sensors=sensors)

    logger.info('writing imported data ...')
    kapture_to_dir(kapture_dir_path, imported_kapture)


def import_7scenes_command_line() -> None:
    """
    Imports RGB-D Dataset 7-Scenes dataset and save them as kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(
        description='Imports RGB-D Dataset 7-Scenes files to the kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True,
                        help='input path Dataset 7-Scenes sequence root path')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('-o', '--output', required=True, help='output directory.')
    parser.add_argument('-p', '--partition', default=None, choices=['mapping', 'query'],
                        help='limit to mapping or query sequences only (using authors split files).')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_7scenes(d7scenes_path=args.input,
                   kapture_dir_path=args.output,
                   force_overwrite_existing=args.force,
                   images_import_method=args.image_transfer,
                   partition=args.partition)


if __name__ == '__main__':
    import_7scenes_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Colmap database import as basic kapture objects functions
"""

from kapture.io.tar import TarCollection
import logging
import numpy as np
from tqdm import tqdm
from typing import Tuple, Optional

# kapture
import kapture
import kapture.io.features
# local
from .database import COLMAPDatabase, blob_to_array, pair_id_to_image_ids
from .database_extra import exists_table
from .cameras import CAMERA_MODEL_NAMES, get_camera_kapture_id_from_colmap_id

logger = logging.getLogger('colmap')


def get_cameras_from_database(database: COLMAPDatabase) -> kapture.Sensors:
    """
    Creates kapture sensors from the colmap database.

    :param database: colmap database
    :return: kapture sensors
    """
    logger.info('parsing cameras  ...')
    kapture_cameras = kapture.Sensors()

    for camera_id, model_id, width, height, params, prior_focal_length in database.execute(
            'SELECT camera_id, model, width, height, params, prior_focal_length FROM cameras;'):
        if model_id not in CAMERA_MODEL_NAMES:
            logger.warning(f'unable to convert colmap camera model ({model_id}) for camera {camera_id}.')
            # use 0 as default
            model_id = 0

        camera_id = get_camera_kapture_id_from_colmap_id(camera_id)
        model_name = CAMERA_MODEL_NAMES[model_id]

        #  By setting the prior_focal_length flag to 0 or 1,
        #  you can give a hint whether the reconstruction algorithm should trust the focal length value.
        params = blob_to_array(params, np.float64)
        params = [width, height] + params.tolist()

        kapture_camera = kapture.Camera(model_name, params)
        kapture_cameras[camera_id] = kapture_camera
    return kapture_cameras


def get_images_and_trajectories_from_database(database: COLMAPDatabase
                                              ) -> Tuple[kapture.RecordsCamera, kapture.Trajectories]:
    """
    Creates records_camera and trajectories from colmap images table
    In trajectories, timestamps are made up from colmap image id.

    :param database: colmap database
    :return: kapture records_camera and trajectories
    """
    logging.info('parsing images ...')
    kapture_images = kapture.RecordsCamera()
    kapture_trajectories = kapture.Trajectories()
    hide_progressbar = logger.getEffectiveLevel() > logging.INFO
    for image_id, name, camera_id, prior_qw, prior_qx, prior_qy, prior_qz, prior_tx, prior_ty, prior_tz \
            in tqdm(database.execute('SELECT image_id, name, camera_id, '
                                     'prior_qw, prior_qx, prior_qy, prior_qz, '
                                     'prior_tx, prior_ty, prior_tz  FROM images;'
                                     ), disable=hide_progressbar):
        # images
        timestamp = int(image_id)
        camera_id = get_camera_kapture_id_from_colmap_id(camera_id)
        kapture_images[timestamp, camera_id] = name
        # trajectories
        prior_q = [prior_qw, prior_qx, prior_qy, prior_qz]
        prior_t = [prior_tx, prior_ty, prior_tz]
        # do not register the pose part if its invalid.
        is_undefined = all(v is None for v in prior_q + prior_t)
        if is_undefined:
            # just ignore
            continue
        prior_pose = kapture.PoseTransform(prior_q, prior_t)
        kapture_trajectories[timestamp, camera_id] = prior_pose

    if len(kapture_trajectories) == 0:
        # if there is no pose at all, just don't bother.
        kapture_trajectories = None

    return kapture_images, kapture_trajectories


def get_keypoints_from_database(database: COLMAPDatabase,
                                records_camera: kapture.RecordsCamera,
                                kapture_dirpath: str,
                                tar_handlers: Optional[TarCollection] = None,
                                keypoints_type: str = 'SIFT'
                                ) -> Optional[kapture.Keypoints]:
    """
    Writes keypoints files and return the kapture keypoints from the colmap database.
    Requires records_camera timestamp == colmap_image_id

    :param database: colmap database.
    :param records_camera: input images.
    :param kapture_dirpath: input root path to kapture.
    :param tar_handlers: collection of preloaded tar archives
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
     (by default, in colmap, it is SIFT, but can be imported)
    :return: kapture keypoints
    """
    image_filenames = set()
    dtype = np.float32
    dsize = None  # usually 6, will be retrieved on first keypoints of DB

    empty_keypoints = []

    # DB query
    colmap_keypoints_request = (
        (image_id, blob_to_array(data, dtype, (rows, cols)) if (rows > 0 and cols > 0) else None)
        for image_id, rows, cols, data in database.execute("SELECT image_id, rows, cols, data FROM keypoints"))

    hide_progressbar = logger.getEffectiveLevel() > logging.INFO
    for colmap_image_id, image_keypoints in tqdm(colmap_keypoints_request, disable=hide_progressbar):
        # retrieve image path from image_id
        assert len(records_camera[colmap_image_id]) == 1
        timestamp = colmap_image_id
        image_filename = next((v for v in records_camera[timestamp].values()), None)
        assert image_filename
        keypoints_filepath = kapture.io.features.get_keypoints_fullpath(keypoints_type,
                                                                        kapture_dirpath,
                                                                        image_filename,
                                                                        tar_handlers)

        # handle no keypoints (delay)
        if image_keypoints is None or image_keypoints.shape[0] == 0:
            logger.warning(f'image={image_filename} has 0 keypoints')
            empty_keypoints.append((image_filename, keypoints_filepath))
            continue

        if image_keypoints.dtype != dtype:
            raise ValueError('inconsistent keypoints type.')
        if dsize is None:
            dsize = int(image_keypoints.shape[1])
        elif dsize != image_keypoints.shape[1]:
            raise ValueError('inconsistent keypoints size.')

        # save the actual file
        kapture.io.features.image_keypoints_to_file(keypoints_filepath, image_keypoints)
        # register it into kapture
        image_filenames.add(image_filename)

    # do all the empty keypoints at the end when we know for sure dsize
    for image_filename, keypoints_filepath in empty_keypoints:
        if dsize is None:
            dsize = 6
        image_keypoints = np.zeros((0, dsize), dtype=dtype)
        # save the actual file
        kapture.io.features.image_keypoints_to_file(keypoints_filepath, image_keypoints)
        # register it into kapture
        image_filenames.add(image_filename)

    if image_filenames:
        return kapture.Keypoints(keypoints_type, dtype, dsize, image_filenames)
    else:
        return None


def get_descriptors_from_database(database: COLMAPDatabase,
                                  images: kapture.RecordsCamera,
                                  kapture_dirpath: str,
                                  tar_handlers: Optional[TarCollection] = None,
                                  keypoints_type: str = 'SIFT',
                                  descriptors_type: str = 'SIFT'
                                  ) -> Optional[kapture.Descriptors]:
    """
    Writes descriptors files and return the list in kapture format from the colmap database.

    :param database: colmap database.
    :param images: list of images (as RecordsCamera).
    :param kapture_dirpath: input root path to kapture.
    :param tar_handlers: collection of preloaded tar archives
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
     (by default, in colmap, it is SIFT, but can be imported)
    :param descriptors_type: type of descriptors to export, name of the descriptors subfolder
     (by default, in colmap, it is SIFT, but can be imported)
    :return: kapture descriptors
    """
    image_filenames = set()
    dtype = np.uint8  # values in the range 0…255
    # see https://colmap.github.io/tutorial.html#feature-detection-and-extraction
    dsize = None  # usually uint8, 128, will be retrieved on first descriptor of DB

    empty_descriptors = []

    colmap_descriptors = (
        (image_id, blob_to_array(data, dtype, (rows, cols)) if (rows > 0 and cols > 0) else None)
        for image_id, rows, cols, data in database.execute("SELECT image_id, rows, cols, data FROM descriptors"))
    hide_progressbar = logger.getEffectiveLevel() > logging.INFO
    for image_id, image_descriptors in tqdm(colmap_descriptors, disable=hide_progressbar):
        # retrieve image path from image_id (actually the timestamp)
        image_filename = next((v for v in images[image_id].values()), None)
        assert image_filename
        descriptors_filepath = kapture.io.features.get_descriptors_fullpath(descriptors_type,
                                                                            kapture_dirpath,
                                                                            image_filename,
                                                                            tar_handlers)

        if image_descriptors is None or image_descriptors.shape[0] == 0:
            logger.warning(f'image={image_id}:{image_filename} has 0 descriptors.')
            empty_descriptors.append((image_filename, descriptors_filepath))
            continue

        if image_descriptors.dtype != dtype:
            raise ValueError('inconsistent descriptors type.')
        if dsize is None:
            dsize = int(image_descriptors.shape[1])
        elif dsize != image_descriptors.shape[1]:
            raise ValueError('inconsistent descriptors size.')

        # save the actual file
        kapture.io.features.image_descriptors_to_file(descriptors_filepath, image_descriptors)
        # register it into to kapture
        image_filenames.add(image_filename)

    # do all the empty descriptors at the end when we know for sure dsize
    for image_filename, descriptors_filepath in empty_descriptors:
        if dsize is None:
            dsize = 128
        image_descriptors = np.zeros((0, dsize), dtype=dtype)
        # save the actual file
        kapture.io.features.image_descriptors_to_file(descriptors_filepath, image_descriptors)
        # register it into kapture
        image_filenames.add(image_filename)

    if image_filenames:
        return kapture.Descriptors(descriptors_type, dtype, dsize, keypoints_type, "L2", image_filenames)
    else:
        return None


def get_matches_from_database(database: COLMAPDatabase,
                              images: kapture.RecordsCamera,
                              kapture_dirpath: str,
                              tar_handlers: Optional[TarCollection] = None,
                              keypoints_type: str = 'SIFT',
                              no_geometric_filtering: bool = False) -> kapture.Matches:
    """
    Writes Matches files and return the list in kapture format from the colmap database.

    :param database: input colmap database.
    :param images: input list of images (as RecordsCamera).
    :param kapture_dirpath: input root path to kapture.
    :param tar_handlers: collection of preloaded tar archives
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    (by default, in colmap, it is SIFT, but can be imported)
    :param no_geometric_filtering: only retrieve matches with geometric consistency.
    :return: kapture matches
    """
    kapture_matches = kapture.Matches()
    # check there is geometric matches available
    matches_table_name = 'matches'
    if not no_geometric_filtering:
        if not exists_table('two_view_geometries', database):
            logger.warning("No table: two_view_geometries: skipping geometric filtering")
        else:
            request = database.execute(
                'SELECT COUNT (*) FROM two_view_geometries')
            nb_verified_matches = next(request)[0]
            if nb_verified_matches > 0:
                request = database.execute('SELECT COUNT (*) FROM matches')
                nb_total_matches = next(request)[0]
                logger.info('keeps {}% of verified matches ({}/{}) ...'.format(
                    nb_verified_matches / nb_total_matches * 100, nb_verified_matches, nb_total_matches))
                matches_table_name = 'two_view_geometries'

    colmap_matches = [(pair_id_to_image_ids(pair_id), blob_to_array(data, np.uint32, (rows, cols)))
                      if rows > 0 else (pair_id_to_image_ids(pair_id), np.empty((rows, cols), dtype=np.uint32))
                      for pair_id, rows, cols, data
                      in database.execute('SELECT pair_id, rows, cols, data FROM {}'.format(matches_table_name))]

    logger.debug('matches: {}'.format(len(colmap_matches)))
    hide_progressbar = logger.getEffectiveLevel() > logging.INFO
    for (image_id1, image_id2), image_matches in tqdm(colmap_matches, disable=hide_progressbar):
        if image_id1 not in images or image_id2 not in images:
            logger.critical('inconsistent image ID {} or {}'.format(image_id1, image_id2))
            continue
        filename1 = next((v for v in images[image_id1].values()), None)
        filename2 = next((v for v in images[image_id2].values()), None)
        assert filename1 and filename2
        if (filename1, filename2) != kapture.Matches.lexical_order(filename1, filename2):
            # have to swap matches (keypoint image1, keypoint image2) become (keypoint image2, keypoint image1)
            image_matches = image_matches[:, ::-1]
            filename1, filename2 = kapture.Matches.lexical_order(filename1, filename2)

        # actually write the file
        # convert colmap image matches into kapture (cast to float and add a score column)
        image_matches = image_matches.astype(np.float)
        image_matches = np.hstack([image_matches, np.zeros((image_matches.shape[0], 1))])
        image_matches_filepath = kapture.io.features.get_matches_fullpath((filename1, filename2),
                                                                          keypoints_type,
                                                                          kapture_dirpath,
                                                                          tar_handlers)
        kapture.io.features.image_matches_to_file(image_matches_filepath, image_matches)
        # register the matching in kapture
        kapture_matches.add(filename1, filename2)

    return kapture_matches



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

import sys
import os.path as path
# when developing, prefer local kapture to the one installed on the system
HERE_PATH = path.normpath(path.dirname(__file__))
REPO_ROOT_PATH = path.normpath(path.dirname(HERE_PATH))
# check the presence of kapture directory in repo to be sure its not the installed version
if path.isdir(path.join(REPO_ROOT_PATH, 'kapture')):
    # workaround for sibling import
    sys.path.insert(0, REPO_ROOT_PATH)



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to import the Scape-Imperial Localisation Dataset dataset to the kapture format.

See research page at https://research.scape.io/silda/
and dataset description there: https://github.com/abmmusa/silda.
"""

import argparse
import logging
import re
import os
import os.path as path
import numpy as np
import quaternion
from typing import Any, Dict, Optional
from tqdm import tqdm
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
import kapture.io.csv
import kapture.io.structure
from kapture.io.binary import TransferAction, transfer_files_from_dir
from kapture.io.records import get_image_fullpath
from kapture.utils.paths import path_secure
import kapture.io.features

logger = logging.getLogger('silda')
SILDA_IMAGE_NAME_PATTERN = re.compile(r'(?P<filename>(?P<timestamp>\d+)_(?P<cam_id>\d+)\.png)')

SILDA_IMAGE_SIZE = np.array([1024, 1024])
SILDA_CALIB_IMAGE_SIZE = np.array([2496, 2496])
SILDA_CORPUS_SPLIT_FILENAMES = {
    'mapping': 'train_imgs.txt',
    'query': 'query_imgs.txt',
}


def _import_cameras(silda_dir_path, snapshots, fallback_cam_model) -> kapture.Sensors:
    logger.info('Processing sensors ...')
    cameras = kapture.Sensors()
    # use hard coded intrinsics
    # evaluated using colmap
    # 1 OPENCV_FISHEYE 1024 1024 393.299 394.815 512 512 -0.223483 0.117325 -0.0326138 0.00361082
    #                  fx, fy, cx, cy, omega
    # 1 FOV 1024 1024 300 300 512 512 0.899632
    cam_id_list = sorted(set(cam_id for _, cam_id, _ in kapture.flatten(snapshots)))
    for cam_id in cam_id_list:
        # pick a image for that cam id
        random_image_intrinsic = next(f'{timestamp}_{cam_id}.intrinsics'  # keep only filename (thats what silda expect)
                                      for timestamp, cid, filename in kapture.flatten(snapshots)
                                      if cid == cam_id)
        logger.debug(f'camera {cam_id} intrinsics : picking at random: ("{random_image_intrinsic}")')
        intrinsic_filepath = path.join(silda_dir_path, 'camera-intrinsics', random_image_intrinsic)
        logger.debug(f'loading file: "{intrinsic_filepath}"')
        silda_proj_params = np.loadtxt(intrinsic_filepath)
        # only retrieve principal point from intrinsics,
        # because the rest correspond to a fisheye model not available in colmap.
        principal_point = (silda_proj_params[0:2] * SILDA_IMAGE_SIZE).flatten().tolist()
        projection = fallback_cam_model
        if 'OPENCV_FISHEYE' == projection:
            focal_length = [393.299, 394.815]
            fisheye_coefficients = [-0.223483,
                                    0.117325, -0.0326138, 0.00361082]
            #          //    fx, fy, cx, cy, k1, k2, k3, k4
            proj_params = focal_length + principal_point + fisheye_coefficients
        elif 'FOV' == projection:
            # use hard coded intrinsics from Torsten reconstruction, ie. :
            #       217.294036, 217.214703, 512.000000, 507.897400, -0.769113
            focal_length = [217.294036, 217.214703]
            # principal_point = [512.000000, 507.897400]
            omega = [-0.769113]
            #                  fx, fy, cx, cy, omega
            proj_params = focal_length + principal_point + omega
        else:
            raise ValueError('Only accepts OPENCV_FISHEYE, or FOV as projection model.')

        camera = kapture.Camera(projection, SILDA_IMAGE_SIZE.tolist() + proj_params)
        cameras[cam_id] = camera
    return cameras


def _import_trajectories(silda_dir_path, image_name_to_ids, hide_progress_bars) -> kapture.Trajectories:
    logger.info('Processing trajectories ...')
    trajectories = kapture.Trajectories()
    with open(path.join(silda_dir_path, 'silda-train-poses.txt')) as file:
        lines = file.readlines()
        lines = (line.rstrip().split() for line in lines)
        extrinsics = {
            line[0]: np.array(line[1:8], dtype=np.float) for line in lines
        }
    for silda_image_name, pose_params in tqdm(extrinsics.items(), disable=hide_progress_bars):
        # Silda poses are 7-dim vectors with the rotation quaternion,
        # and the translation vector. The order needs to be:
        # qw,qx,qy,qz,tx,ty,tz
        # The parameters should be described in terms of camera to world transformations
        if silda_image_name not in image_name_to_ids:
            # if this is not referenced: means its part of the corpus to be ignored.
            continue
        pose = kapture.PoseTransform(pose_params[0:4], pose_params[4:7]).inverse()
        timestamp, cam_id = image_name_to_ids[silda_image_name]
        trajectories[timestamp, cam_id] = pose
    # if query, trajectories is empty, so juste do not save it
    if len(trajectories) == 0:
        trajectories = None
    return trajectories


def _make_rigs(replace_pose_rig, trajectories) -> kapture.Rigs:
    logger.info('Making up a rig ...')
    rigs = kapture.Rigs()
    pose_babord = kapture.PoseTransform(t=[0, 0, 0], r=quaternion.from_rotation_vector([0, -np.pi / 2, 0]))
    pose_tribord = kapture.PoseTransform(t=[0, 0, 0], r=quaternion.from_rotation_vector([0, np.pi / 2, 0]))
    rigs['silda_rig', '0'] = pose_babord
    rigs['silda_rig', '1'] = pose_tribord
    if replace_pose_rig:
        logger.info('replacing camera poses with rig poses.')
        kapture.rigs_recover_inplace(trajectories, rigs)
    return rigs


def import_silda(silda_dir_path: str,
                 destination_kapture_dir_path: str,
                 fallback_cam_model: str = 'FOV',
                 do_split_cams: bool = False,
                 corpus: Optional[str] = None,
                 replace_pose_rig: bool = False,
                 force_overwrite_existing: bool = False,
                 images_import_strategy: TransferAction = TransferAction.link_absolute) -> None:
    """
    Imports data from silda dataset.

    :param silda_dir_path: path to the silda top directory
    :param destination_kapture_dir_path: input path to kapture directory.
    :param fallback_cam_model: camera model to fallback when necessary
    :param do_split_cams: If true, re-organises and renames the image files to split apart cameras.
    :param corpus: the list of corpus to be imported, among 'mapping', 'query'.
    :param replace_pose_rig: if True, replaces poses of individual cameras with poses of the rig.
    :param force_overwrite_existing: if true, Silently overwrite kapture files if already exists.
    :param images_import_strategy: how to copy image files.
    """

    # sanity check
    silda_dir_path = path_secure(path.abspath(silda_dir_path))
    destination_kapture_dir_path = path_secure(path.abspath(destination_kapture_dir_path))
    if TransferAction.root_link == images_import_strategy and do_split_cams:
        raise ValueError('impossible to only link images directory and applying split cam.')
    hide_progress_bars = logger.getEffectiveLevel() >= logging.INFO

    # prepare output directory
    kapture.io.structure.delete_existing_kapture_files(destination_kapture_dir_path, force_overwrite_existing)
    os.makedirs(destination_kapture_dir_path, exist_ok=True)

    # images ###########################################################################################################
    logger.info('Processing images ...')
    # silda-images
    #   ...
    #   ├── 1445_0.png
    #   ├── 1445_1.png
    #   ...
    silda_images_root_path = path.join(silda_dir_path, 'silda-images')
    # list all png files (its PNG in silda) using a generator.
    if corpus is not None:
        assert corpus in SILDA_CORPUS_SPLIT_FILENAMES
        # if corpus specified, filter by those which directory name match corpus.
        logger.debug(f'only importing {corpus} part.')
        corpus_file_path = path.join(silda_dir_path, SILDA_CORPUS_SPLIT_FILENAMES[corpus])
        with open(corpus_file_path, 'rt') as corpus_file:
            corpus_filenames = corpus_file.readlines()
            image_filenames_original = sorted(
                filename.strip()
                for filename in corpus_filenames
            )
    else:
        image_filenames_original = sorted(
            filename
            for dir_path, sd, fs in os.walk(silda_images_root_path)
            for filename in fs
            if filename.endswith('.png'))

    image_filenames_kapture = []
    snapshots = kapture.RecordsCamera()
    image_name_to_ids = {}  # '1445_0.png' -> (1445, 0)
    for image_filename_original in tqdm(image_filenames_original, disable=hide_progress_bars):
        # retrieve info from image filename
        name_parts_match = SILDA_IMAGE_NAME_PATTERN.match(image_filename_original)
        assert name_parts_match is not None
        shot_info: Dict[str, Any]
        shot_info = name_parts_match.groupdict()
        shot_info['timestamp'] = int(shot_info['timestamp'])  # To avoid warnings about type of the value
        # eg. file_info = {'filename': '1445_0.png', 'timestamp': 1445, 'cam_id': '0'}
        # create a path of the image into NLE dir
        if do_split_cams:
            # re-organise images with subfolders per corpus/camera/timestamp.png
            kapture_image_filename = path.join(shot_info['cam_id'],
                                               '{:04d}.png'.format(shot_info['timestamp']))
        else:
            # keep the original file hierarchy
            kapture_image_filename = image_filename_original

        image_filenames_kapture.append(kapture_image_filename)
        snapshots[shot_info['timestamp'], shot_info['cam_id']] = kapture_image_filename
        image_name_to_ids[shot_info['filename']] = (shot_info['timestamp'], shot_info['cam_id'])

    assert len(image_filenames_kapture) == len(image_filenames_original)
    # intrinsics #######################################################################################################
    cameras = _import_cameras(silda_dir_path, snapshots, fallback_cam_model)

    # extrinsics #######################################################################################################
    trajectories = _import_trajectories(silda_dir_path, image_name_to_ids, hide_progress_bars)

    # rigs
    rigs = _make_rigs(replace_pose_rig, trajectories)

    # pack it all together
    kapture_data = kapture.Kapture(
        sensors=cameras,
        records_camera=snapshots,
        trajectories=trajectories,
        rigs=rigs
    )

    logger.info('saving to Kapture  ...')
    kapture.io.csv.kapture_to_dir(destination_kapture_dir_path, kapture_data)

    # finally import images
    if images_import_strategy != TransferAction.skip:
        # importing image files
        logger.info(f'importing {len(image_filenames_original)} images ...')
        assert len(image_filenames_original) == len(image_filenames_kapture)
        image_file_paths_original = [
            path.join(silda_images_root_path, image_filename_kapture)
            for image_filename_kapture in image_filenames_original]
        image_file_paths_kapture = [
            get_image_fullpath(destination_kapture_dir_path, image_filename_kapture)
            for image_filename_kapture in image_filenames_kapture]
        transfer_files_from_dir(image_file_paths_original, image_file_paths_kapture, images_import_strategy)
    logger.info('done.')


def import_silda_command_line():
    """
    Do the SILDa to kapture import using the command line parameters provided by the user.
    """
    parser = argparse.ArgumentParser(description='imports SILDa dataset to kapture format.')
    ####################################################################################################################
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete kapture if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input',
                        help='path to silda root directory.')
    parser.add_argument('-o', '--output', required=True,
                        help='output directory where to save NLE files.')
    parser.add_argument('-s', '--split_cams', action='store_true', default=False,
                        help='reorganises the image file per camera folders.')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('--corpus', choices=['mapping', 'query'],
                        help='restrain (or not) do only mapping or query images.')
    parser.add_argument('--cam_model', choices=['OPENCV_FISHEYE', 'FOV'], default='FOV',
                        help='camera model to be used.')
    parser.add_argument('--rig_collapse', action='store_true', default=False,
                        help='Replace camera poses with rig poses.')
    ####################################################################################################################
    args = parser.parse_args()
    logger.setLevel(args.verbose)

    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_silda(args.input,
                 args.output,
                 fallback_cam_model=args.cam_model,
                 do_split_cams=args.split_cams,
                 corpus=args.corpus,
                 replace_pose_rig=args.rig_collapse,
                 force_overwrite_existing=args.force,
                 images_import_strategy=args.image_transfer)


if __name__ == '__main__':
    import_silda_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Rename features inside kapture data
"""

import logging
import os
import os.path as path
import argparse
import shutil

import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
import kapture.io.features
import kapture.io.csv

logger = logging.getLogger('rename_features')


def rename_features(kapture_dirpath: str,
                    feature_type: str,
                    current_name: str,
                    new_name: str) -> None:
    """
    Rename features inside kapture data.
    """
    with kapture.io.csv.get_all_tar_handlers(kapture_dirpath) as tar_handlers:
        kapture_data = kapture.io.csv.kapture_from_dir(kapture_dirpath, tar_handlers=tar_handlers)
    if feature_type == 'global_features':
        # rename global_features folder
        if kapture_data.global_features is None or current_name not in kapture_data.global_features:
            raise ValueError(f'couldn\'t find {feature_type} in '
                             f'kapture_data.global_features: {kapture_data.global_features}')
        inpath = kapture.io.features.get_features_fullpath(kapture.GlobalFeatures, current_name, kapture_dirpath)
        outpath = kapture.io.features.get_features_fullpath(kapture.GlobalFeatures, new_name, kapture_dirpath)
        if path.isdir(outpath):
            raise ValueError(f'{outpath} already exists')
        os.rename(inpath, outpath)
    elif feature_type == 'descriptors':
        # rename descriptors folder
        if kapture_data.descriptors is None or current_name not in kapture_data.descriptors:
            raise ValueError(f'couldn\'t find {feature_type} in '
                             f'kapture_data.descriptors: {kapture_data.descriptors}')
        inpath = kapture.io.features.get_features_fullpath(kapture.Descriptors, current_name, kapture_dirpath)
        outpath = kapture.io.features.get_features_fullpath(kapture.Descriptors, new_name, kapture_dirpath)
        if path.isdir(outpath):
            raise ValueError(f'{outpath} already exists')
        os.rename(inpath, outpath)
    elif feature_type == 'keypoints':
        # rename keypoints folder
        if kapture_data.keypoints is not None and current_name in kapture_data.keypoints:
            inpath = kapture.io.features.get_features_fullpath(kapture.Keypoints, current_name, kapture_dirpath)
            outpath = kapture.io.features.get_features_fullpath(kapture.Keypoints, new_name, kapture_dirpath)
            if path.isdir(outpath):
                raise ValueError(f'{outpath} already exists')
            os.rename(inpath, outpath)
        else:
            logger.warning(f'couldn\'t find {feature_type} in '
                           f'kapture_data.keypoints: {kapture_data.keypoints}')

        # rename matches folder
        if kapture_data.matches is not None and current_name in kapture_data.matches:
            inpath = kapture.io.features.get_features_fullpath(kapture.Matches, current_name, kapture_dirpath)
            outpath = kapture.io.features.get_features_fullpath(kapture.Matches, new_name, kapture_dirpath)
            if path.isdir(outpath):
                raise ValueError(f'{outpath} already exists')
            os.rename(inpath, outpath)
        else:
            logger.debug(f'couldn\'t find {feature_type} in '
                         f'kapture_data.matches: {kapture_data.matches}')

        # update descriptors .txt files that uses this keypoints
        if kapture_data.descriptors is not None:
            for descriptors_type, kapture_descriptors in kapture_data.descriptors.items():
                if kapture_descriptors.keypoints_type == current_name:
                    logger.debug(f'descriptors {descriptors_type}, changing keypoints_type')
                    descriptors_csv_path = kapture.io.csv.get_feature_csv_fullpath(kapture.Descriptors,
                                                                                   descriptors_type,
                                                                                   kapture_dirpath)
                    out_kapture_descriptors = kapture.Descriptors(kapture_descriptors.type_name,
                                                                  kapture_descriptors.dtype,
                                                                  kapture_descriptors.dsize,
                                                                  new_name,
                                                                  kapture_descriptors.metric_type,
                                                                  kapture_descriptors)
                    kapture.io.csv.descriptors_to_file(descriptors_csv_path, out_kapture_descriptors)

        # update observations.txt
        if kapture_data.observations is not None:
            if kapture_data.keypoints is None or current_name not in kapture_data.keypoints:
                raise ValueError(f'{current_name} not found in keypoints, perhaps a previous attempt failed midway'
                                 'observations won\'t load properly in this state so fix them manually')
            new_observations = kapture.Observations()
            for point3d_idx, keypoints_type in kapture_data.observations.key_pairs():
                output_keypoints_type = keypoints_type
                if output_keypoints_type == current_name:
                    output_keypoints_type = new_name
                observation_array = kapture_data.observations.get(point3d_idx)[keypoints_type]
                new_observations.setdefault(point3d_idx, {})[output_keypoints_type] = observation_array
            observations_csv_path = kapture.io.csv.get_csv_fullpath(kapture.Observations, kapture_dirpath)
            observations_csv_path_bak = observations_csv_path + '.rename_bak'
            if path.isfile(observations_csv_path_bak):
                os.remove(observations_csv_path_bak)
            shutil.copy(observations_csv_path, observations_csv_path_bak)
            kapture.io.csv.observations_to_file(observations_csv_path, new_observations)
    else:
        raise ValueError(f'feature_type {feature_type} is unknown')
    logger.debug('all done!')


def rename_features_command_line() -> None:
    """
    Rename features inside kapture data.
    """
    parser = argparse.ArgumentParser(
        description='rename features inside kapture data')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # export ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to kapture directory')
    parser.add_argument('--feature-type', required=True, choices=['keypoints', 'descriptors', 'global_features'],
                        help='types of features to rename.')
    parser.add_argument('--current-name', required=True,
                        help='keypoints_type, descriptors_type, global_features_type depending on selected features.')
    parser.add_argument('--new-name', required=True,
                        help='keypoints_type, descriptors_type, global_features_type depending on selected features.')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    rename_features(args.input, args.feature_type, args.current_name, args.new_name)


if __name__ == '__main__':
    rename_features_command_line()



# Copyright 2021-present NAVER Corp. Under BSD 3-clause license

"""
Upgrade operations.
"""

import os
import os.path as path
from typing import Optional
import shutil
# import numpy as np like in kapture.io.csv
# so that types written as "np.float32" are understood by read_old_image_features_csv
import numpy as np  # noqa: F401

import kapture
import kapture.utils.logging
import kapture.io.features
import kapture.io.csv
from kapture.utils.paths import populate_files_in_dirpath
from kapture.utils.logging import getLogger

CSV_FILENAMES_1_0 = [
    path.join('sensors', 'sensors.txt'),
    path.join('sensors', 'trajectories.txt'),
    path.join('sensors', 'rigs.txt'),
    path.join('sensors', 'records_camera.txt'),
    path.join('sensors', 'records_depth.txt'),
    path.join('sensors', 'records_lidar.txt'),
    path.join('sensors', 'records_wifi.txt'),
    path.join('sensors', 'records_bluetooth.txt'),
    path.join('sensors', 'records_gnss.txt'),
    path.join('sensors', 'records_accelerometer.txt'),
    path.join('sensors', 'records_gyroscope.txt'),
    path.join('sensors', 'records_magnetic.txt'),
    path.join('reconstruction', 'points3d.txt')]


def read_old_image_features_csv(csv_filepath: str):
    """
    Read the old image feature

    :param csv_filepath: the path to the csv file containing image features
    """
    with open(csv_filepath, 'r') as source_file:
        table = kapture.io.csv.table_from_file(source_file)
        line = list(table)[0]
        assert len(line) == 3
        name, dtype, dsize = line[0], line[1], int(line[2])

    # try to list all possible type from numpy that can be used in eval(dtype)
    from numpy import float, float32, float64, int32, uint8  # noqa: F401
    if isinstance(type(eval(dtype)), type):
        dtype = eval(dtype)
    else:
        raise ValueError('Expect data type ')
    return name, dtype, dsize


def upgrade_1_0_to_1_1_inplace(kapture_dirpath: str,  # noqa: C901: function a bit long but well documented
                               keypoints_type: Optional[str],
                               descriptors_type: Optional[str],
                               global_features_type: Optional[str],
                               descriptors_metric_type: str,
                               global_features_metric_type: str) -> None:
    """
    Do the upgrade from 1.0 to 1.1 version in place: will replace all the necessary files.

    """
    # some text files didn't change, just change their header
    for csv_filename in CSV_FILENAMES_1_0:
        csv_fullpath = path.join(kapture_dirpath, csv_filename)
        if path.isfile(csv_fullpath):
            getLogger().debug(f'converting {csv_fullpath}...')
            old_version = kapture.io.csv.get_version_from_csv_file(csv_fullpath)
            assert (old_version is None or old_version == '1.0')
            with open(csv_fullpath, 'r') as source_file:
                if old_version is not None:
                    source_file.readline()  # read and ignore header
                # write replacement header
                lines = source_file.read()
            with open(csv_fullpath, 'w') as source_file:
                lines_all = kapture.io.csv.KAPTURE_FORMAT_1 + kapture.io.csv.kapture_linesep + lines
                source_file.write(lines_all)
    # keypoints
    keypoints_dir_path = path.join(kapture_dirpath, 'reconstruction', 'keypoints')
    keypoints_csv_path = path.join(keypoints_dir_path, 'keypoints.txt')
    local_features_json_filepath = path.join(keypoints_dir_path, 'extract_local_features.json')
    if path.isdir(keypoints_dir_path) and path.isfile(keypoints_csv_path):
        getLogger().debug(f'converting {keypoints_dir_path}...')
        old_version = kapture.io.csv.get_version_from_csv_file(keypoints_csv_path)
        assert old_version is None or old_version == '1.0'
        name, dtype, dsize = read_old_image_features_csv(keypoints_csv_path)
        os.remove(keypoints_csv_path)
        if keypoints_type is None:
            assert name != ''
            keypoints_type = name
        keypoints = kapture.Keypoints(name, dtype, dsize)
        keypoints_csv_output_path = path.join(kapture_dirpath,
                                              kapture.io.csv.FEATURES_CSV_FILENAMES[kapture.Keypoints](keypoints_type))
        keypoints_output_dir = path.dirname(keypoints_csv_output_path)
        kapture.io.csv.keypoints_to_file(keypoints_csv_output_path, keypoints)

        # json file
        if path.isfile(local_features_json_filepath):
            local_features_json_output_file = path.join(keypoints_output_dir, 'extract_local_features.json')
            shutil.move(local_features_json_filepath, local_features_json_output_file)

        # now copy all .kpt files
        keypoints_filenames = list(populate_files_in_dirpath(keypoints_dir_path, '.kpt'))
        # cast to list before enumerating (or moved files will be listed multiple times)
        for keypoints_filename in keypoints_filenames:
            keypoints_output_file = path.join(keypoints_output_dir, keypoints_filename)
            keypoints_inpath = path.join(keypoints_dir_path, keypoints_filename)
            os.makedirs(path.dirname(keypoints_output_file), exist_ok=True)
            shutil.move(keypoints_inpath, keypoints_output_file)
            old_kp_dir = path.dirname(keypoints_inpath)
            if len(os.listdir(old_kp_dir)) == 0:
                os.removedirs(old_kp_dir)

    # descriptors
    descriptors_dir_path = path.join(kapture_dirpath, 'reconstruction', 'descriptors')
    descriptors_csv_path = path.join(descriptors_dir_path, 'descriptors.txt')
    if path.isdir(descriptors_dir_path) and path.isfile(descriptors_csv_path):
        getLogger().debug(f'converting {descriptors_dir_path}...')
        old_version = kapture.io.csv.get_version_from_csv_file(descriptors_csv_path)
        assert old_version is None or old_version == '1.0'
        assert keypoints_type is not None
        name, dtype, dsize = read_old_image_features_csv(descriptors_csv_path)
        os.remove(descriptors_csv_path)
        if descriptors_type is None:
            assert name != ''
            descriptors_type = name
        descriptors = kapture.Descriptors(name, dtype, dsize, keypoints_type, descriptors_metric_type)
        descriptors_csv_output_path = path.join(kapture_dirpath,
                                                kapture.io.csv.FEATURES_CSV_FILENAMES[kapture.Descriptors](
                                                    descriptors_type)
                                                )
        descriptors_output_dir = path.dirname(descriptors_csv_output_path)
        kapture.io.csv.descriptors_to_file(descriptors_csv_output_path, descriptors)
        # now copy all .desc files
        descriptors_filenames = list(populate_files_in_dirpath(descriptors_dir_path, '.desc'))
        # cast to list before enumerating (or moved files will be listed multiple times)
        for descriptors_filename in descriptors_filenames:
            descriptors_output_file = path.join(descriptors_output_dir, descriptors_filename)
            os.makedirs(path.dirname(descriptors_output_file), exist_ok=True)
            shutil.move(path.join(descriptors_dir_path, descriptors_filename), descriptors_output_file)
            old_desc_dir = path.dirname(path.join(descriptors_dir_path, descriptors_filename))
            if len(os.listdir(old_desc_dir)) == 0:
                os.removedirs(old_desc_dir)

    # matches
    matches_dir_path = path.join(kapture_dirpath, 'reconstruction', 'matches')
    matches_json_filepath = path.join(matches_dir_path, 'run_matching.json')
    if path.isdir(matches_dir_path):
        getLogger().debug(f'converting {matches_dir_path}...')
        assert keypoints_type is not None
        matches_output_dir = kapture.io.features.get_matches_fullpath(None, keypoints_type, kapture_dirpath)
        os.makedirs(matches_output_dir, exist_ok=True)

        # json file
        if path.isfile(matches_json_filepath):
            matches_json_output_file = path.join(matches_output_dir, 'run_matching.json')
            shutil.move(matches_json_filepath, matches_json_output_file)

        # now copy all .matches files
        matches_filenames = list(populate_files_in_dirpath(matches_dir_path, '.matches'))
        # cast to list before enumerating (or moved files will be listed multiple times)
        for matches_filename in matches_filenames:
            matches_output_file = path.join(matches_output_dir, matches_filename)
            os.makedirs(path.dirname(matches_output_file), exist_ok=True)
            shutil.move(path.join(matches_dir_path, matches_filename), matches_output_file)
            old_matches_dir = path.dirname(path.join(matches_dir_path, matches_filename))
            if len(os.listdir(old_matches_dir)) == 0:
                os.removedirs(old_matches_dir)

    # global features
    global_features_dir_path = path.join(kapture_dirpath, 'reconstruction', 'global_features')
    global_features_csv_path = path.join(global_features_dir_path, 'global_features.txt')
    global_features_json_filepath = path.join(global_features_dir_path, 'extract_global_features.json')
    if path.isdir(global_features_dir_path) and path.isfile(global_features_csv_path):
        getLogger().debug(f'converting {global_features_dir_path}...')
        old_version = kapture.io.csv.get_version_from_csv_file(global_features_csv_path)
        assert old_version is None or old_version == '1.0'
        assert keypoints_type is not None
        name, dtype, dsize = read_old_image_features_csv(global_features_csv_path)
        os.remove(global_features_csv_path)
        if global_features_type is None:
            assert name != ''
            global_features_type = name
        global_features = kapture.GlobalFeatures(name, dtype, dsize, global_features_metric_type)
        global_features_csv_output_path = path.join(kapture_dirpath,
                                                    kapture.io.csv.FEATURES_CSV_FILENAMES[kapture.GlobalFeatures](
                                                        global_features_type)
                                                    )
        global_features_output_dir = path.dirname(global_features_csv_output_path)
        kapture.io.csv.global_features_to_file(global_features_csv_output_path, global_features)

        # json file
        if path.isfile(global_features_json_filepath):
            global_features_json_output_file = path.join(global_features_output_dir, 'extract_global_features.json')
            shutil.move(global_features_json_filepath, global_features_json_output_file)

        # now copy all .gfeat files
        global_features_filenames = list(populate_files_in_dirpath(global_features_dir_path, '.gfeat'))
        # cast to list before enumerating (or moved files will be listed multiple times)
        for global_features_filename in global_features_filenames:
            global_features_output_file = path.join(global_features_output_dir, global_features_filename)
            os.makedirs(path.dirname(global_features_output_file), exist_ok=True)
            shutil.move(path.join(global_features_dir_path, global_features_filename), global_features_output_file)
            old_gfeat_dir = path.dirname(path.join(global_features_dir_path, global_features_filename))
            if len(os.listdir(old_gfeat_dir)) == 0:
                os.removedirs(old_gfeat_dir)

    # observations
    observations_csv_filename = path.join('reconstruction', 'observations.txt')
    observations_csv_path = path.join(kapture_dirpath, observations_csv_filename)
    if path.isfile(observations_csv_path):
        getLogger().debug(f'converting {observations_csv_path}...')
        old_version = kapture.io.csv.get_version_from_csv_file(observations_csv_path)
        assert old_version is None or old_version == '1.0'
        assert keypoints_type is not None
        observations = kapture.Observations()
        with open(observations_csv_path, 'r') as source_file:
            table = kapture.io.csv.table_from_file(source_file)
            # point3d_id, [image_path, feature_id]*
            for points3d_id_str, *pairs in table:
                points3d_id = int(points3d_id_str)
                if len(pairs) > 1:
                    image_paths = pairs[0::2]
                    keypoints_ids = pairs[1::2]
                    for image_path, keypoint_id in zip(image_paths, keypoints_ids):
                        observations.add(points3d_id, keypoints_type, image_path, int(keypoint_id))
        kapture.io.csv.observations_to_file(observations_csv_path, observations)
    getLogger().debug('all done!')
